/*
Copyright 1990, Carnegie-Mellon University. Permission is granted to any
individual to use, or  redistribute this software so long as it is not sold for
profit and provided this copyright is retained.

Carnegie-Mellon may charge a reproduction  fee to  recover costs of media,
packaging, printing, shipping, order processing, or any computer use required
for reproduction.

This software is provided "as is". We do not warrant in any way the performance
of the software included in this distribution, nor the accuracy of any related
documentation, and neither the authors of any of these programs or dcoument-
ation nor Carnegie-Mellon University acknowledge any liability resulting from
program or documentation errors.

This product may  have been built using tools, compilers, and utilities from
other vendors. Where ever possible we have made efforts to include copyright
notices  in our product for these other vendors. It is your responsibilty to
ensure that no copyrights violations exist when you use  our product.

This release is reasonably stable. We are not aware of any major bugs in it.
We developed this software for our environment at Carnegie-Mellon University.
Certain features of this product may not work at other sites.

We would like to hear of problems, bugs or suggestions. Send mail to:
org.special-projects.software-distribution@andrew.cmu.edu

For a detailed  discussion of the requirements and installation of this
product please read the Install document.
*/

/*
 * screen handling module -- Machine dependent
 * curses/termcap interface
 */


#include <stdio.h>
#include <curses.h>
#include <ctype.h>
#include <signal.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <term.h>
#include <sys/ioctl.h>

#undef HZ
#ifdef __hpux
#undef TRUE
#undef FALSE
#endif

#include <curses.h>
#include <ctype.h>
WINDOW *curswin;
#ifdef __SVR4
#define __EXTENSIONS__
#endif
#include <signal.h>

/*#if !defined(__hpux) && !defined(__SVR4) && !defined(__linux__)
#include <sgtty.h>
#endif */

#include <common.h>
#include <and_keys.h>
#include <comkeys.h>
#include <screen.h>
#include <msgs.h>

#if defined(__linux__) && defined(__GLIBC__)
#include <sys/ioctl.h>
#endif

#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
#include <termios.h>
#include <term.h>
#endif

#ifdef __SVR4
#include <ucontext.h>
#include <sys/filio.h>
#include <sys/iso/signal_iso.h>
#endif
#if defined(__hpux) || defined(__SVR4) || defined(ultrix) || defined(__linux__)
extern char *tgetstr(), _putchar();
#define _putchar_rtn _putchar
#define _pfast_thingy _pfast
#else
extern char  _putchar();
#define _putchar_rtn _putchar
#define _pfast_thingy __pfast
#endif
#define DEBUG
#ifdef DEBUG
static FILE *debug_file;
#endif

#include "dialog.h"
#include "windows.h"
#include "and_scr.h"
#include "console.h"

/*
 * hp has new names for the terminal strings
 */

#ifdef cursor_address
#undef _putchar_rtn
#define _putchar_rtn putchar

#undef CM
#define BC cursor_left /*should be backspace if not ^H*/
#define BT back_tab
#define BS (cursor_left!=0) /*should be has back space*/
#define CA (cursor_address!=0) /*should be has cursor addressing*/
#define CM cursor_address
#define CR carriage_return
#define NL "\n"			/*should be new line*/
#define HO cursor_home
#define GT (tab!=0)		/*has hard tabs*/
#define LL cursor_to_ll
#define NC FALSE		/*return doesn't work right*/
#define ND cursor_right
#define SO enter_standout_mode
#define TA tab
#define UP cursor_up

#endif

#define enter_alt_charset_cap "as"
#define exit_alt_charset_cap  "ae"
#define enter_bold_cap        "md"
#define enter_blink_cap       "mb"
#define exit_attribute_cap    "me"
#define init_string_cap       "is"
#define enter_keypad_cap      "ks"
#define exit_keypad_cap       "ke"

#define ATTRSET attrset

static int     cur_mode = NORMAL;

unsigned char Termcap_AttrTable[] = {TERMCAP_ATTR_TABLE};

unsigned char *AttrTablePtr;

static char charmap[256];
static int nostandout;

static unsigned char *refresh_buffer = 0;
static unsigned char *refresh_attr   = 0;

static int cur_char_set = NO_C_SET;

Boolean CursorSaved = FALSE;
unsigned char cursor_row, cursor_col, screen_rows, screen_cols, status_row;
unsigned char saved_row,  saved_col;

#define	MAX_TC_LEN  50

static	char TCkr[MAX_TC_LEN+1];	/* arrow key right */
static	char TCku[MAX_TC_LEN+1];	/* arrow key up */
static	char TCkd[MAX_TC_LEN+1];	/* arrow key down */
static	char TCkl[MAX_TC_LEN+1];	/* arrow key left */
static	char TCkN[MAX_TC_LEN+1];	/* page down */
static	char TCkP[MAX_TC_LEN+1];	/* page up */
static	char TCkh[MAX_TC_LEN+1];	/* home key*/
static	char TCk1[MAX_TC_LEN+1];	/* F1 key */
static	char TCk2[MAX_TC_LEN+1];	/* F2 key */
static	char TCk3[MAX_TC_LEN+1];	/* F3 key */
static	char TCk4[MAX_TC_LEN+1];	/* F4 key */
static	char TCkI[MAX_TC_LEN+1];	/* Insert key */
static	char TCkR[MAX_TC_LEN+1];	/* scroll up */
static	char TCkF[MAX_TC_LEN+1];	/* scroll forward */
static	char TCkE[MAX_TC_LEN+1];	/* Clear to eol -- simulate end key */

static int  PredefinedKeys[] = {
    KEYCODE_CTRL_HOME,
    KEYCODE_PAGE_UP,
    KEYCODE_F1,
    KEYCODE_F2,
    KEYCODE_F3,
    KEYCODE_F4,
    -1
};

typedef struct keypad_struct {
    char *str;
    char *termcap_id;
    int   keycode;
} KEYPAD;

static KEYPAD KeyPad[] = {
    {TCkr, "kr", KEYCODE_RIGHT},
    {TCkl, "kl", KEYCODE_LEFT},
    {TCku, "ku", KEYCODE_UP},
    {TCkd, "kd", KEYCODE_DOWN},
    {TCkN, "kN", KEYCODE_PAGE_DOWN},
    {TCkP, "kP", KEYCODE_PAGE_UP},
    {TCkh, "kh", KEYCODE_HOME},
    {TCk1, "k1", KEYCODE_F1},
    {TCk2, "k2", KEYCODE_F2},
    {TCk3, "k3", KEYCODE_F3},
    {TCk4, "k4", KEYCODE_F4},
    {TCkI, "kI", KEYCODE_INSERT},
    {TCkR, "kR", KEYCODE_PAGE_UP},
    {TCkF, "kF", KEYCODE_PAGE_DOWN},
    {TCkE, "kE", KEYCODE_END},
    {NIL,  NIL,  -1}
};

static int gl_BytesRemaining = 0; /* Number of unread bytes in tty buf */

static char *tc_start_alt_set = NIL;
static char *tc_end_alt_set = NIL;
static char *tc_set_blink = NIL;
static char *tc_set_bold = NIL;
static char *tc_set_all_off = NIL;
static char *tc_cap_ks = NIL;
static char *tc_cap_ke = NIL;

static char CAP_as[MAX_TC_LEN+1] = {0}; /* start alternate char set */
static char CAP_ae[MAX_TC_LEN+1] = {0}; /* end alternate char set */
static char CAP_mb[MAX_TC_LEN+1] = {0}; /* set blink attribute */
static char CAP_md[MAX_TC_LEN+1] = {0}; /* set bold attribute */
static char CAP_me[MAX_TC_LEN+1] = {0}; /* turn off all attributes */
static char CAP_ks[MAX_TC_LEN+1] = {0}; /* into keypad transmit mode */
static char CAP_ke[MAX_TC_LEN+1] = {0}; /* out of keypad transmit mode */

static char TCbuf[1024];

void SCR_SetHighlight();
int wait_keyboard_io(int really_wait);

void SCR_UpdateScreen(void)
{
  refresh();
}

void SCR_Beep (void)
{
  putchar('\07');
  SCR_UpdateScreen();
}

int SCR_Keyboard_io(void)
{
  return((wait_keyboard_io(FALSE) > 0 ||
	  gl_BytesRemaining > 0) ? TRUE : FALSE);
}

int wait_keyboard_io(int really_wait)
{
  fd_set readfds;
  struct timeval tml;
  memset(&readfds,0,sizeof(readfds));
  FD_SET(fileno(stdin),&readfds);
  tml.tv_sec = 0;
  tml.tv_usec = 1;

  if (really_wait && gl_BytesRemaining > 0) return(TRUE);
  
 /*
  * some unixes are strange with a select with less than
  * FD_SETSIZE
  */
  if(really_wait)
     return(select(FD_SETSIZE, (fd_set*)&readfds, 0, 0,0));
  return(select(FD_SETSIZE, (fd_set*)&readfds, 0, 0, &tml));
}

static int bytes_available(void)
{
  if (gl_BytesRemaining <= 0) {
    int arg;
    if (ioctl(0, FIONREAD, &arg) != 0) arg = 0;
    gl_BytesRemaining = arg;
  }
  return(gl_BytesRemaining);
}

int SCR_ClearKeyboard(void)
{
  int user_break = 0;

  while(SCR_Keyboard_io()) {
    if ((SCR_GetKey()) == CONTROL_BREAK) user_break++;
  }
  return(user_break);
}

static int GetKey(void)
{
  int key = getch();

  gl_BytesRemaining = imax(0, (gl_BytesRemaining - 1));

  return(key);
}

int SCR_GetKey(void)
{
    void SCR_WriteStatusLine();
    void SCR_ClearStatusLine();
    static int may_help_with_escape = FALSE;
    int ch, i, nbytes;
    int csl = FALSE;

    refresh();
    wait_keyboard_io(TRUE);
          
    nbytes = bytes_available();

#ifdef ultrix
    ch = GetKey();
#else
    ch = GetKey() & 0xff;
#endif

    if (ch == ((unsigned char)EOF)) {
      if (feof(stdin)) {
	kill(getpid(), SIGTERM); /* EOF on stdin */
      }
    }

    ch &= 0x7f;

    if (ch == ESC_KEY) { /* ESC key */
    next_key:
      ch = GetKey() & 0x7f;
      if (ch == ESC_KEY) {
	if (!may_help_with_escape) {
	  may_help_with_escape = TRUE;
	  return(0);
	}
	while (SCR_Keyboard_io()) SCR_GetKey();
	SCR_WriteStatusLine("ESC - release the ESC key, then press the next key.");
	csl = TRUE;
	goto next_key;
      }
      if (csl) SCR_ClearStatusLine();
      ch += 256;
      i = 0;
      while(PredefinedKeys[i] != -1)
	if (PredefinedKeys[i++] == ch) goto got_a_key;
      ch = CheckKeypad(ch, TRUE);
    } else {
      if (ch == KEYCODE_ALT_SET_MARK) ch = KEYCODE_SET_MARK;
    }
    if (ch == KEYCODE_F10) ch = KEYCODE_ALT_QMARK;

  got_a_key:
    may_help_with_escape = FALSE;
    return(ch);
}

int CheckKeypad(int chin, int prefixed)
{
    int  endbuf, i, ThereisHope = TRUE;
    char chbuf[MAX_TC_LEN+1];

    if (prefixed) {
	chbuf[0] = ESC_KEY;
	chbuf[1] = chin - 256;
	chbuf[2] = '\0';
    } else {
	chbuf[0] = chin;
	chbuf[1] = '\0';
    }

    endbuf = strlen(chbuf);

#ifdef DEBUG
    fprintf ( debug_file, "Input is: %d %d\n", (int)chbuf[0], (int)chbuf[1] );
#endif

    while (endbuf < MAX_TC_LEN) {
	ThereisHope = FALSE;
	i = 0;
	while (KeyPad[i].str != NIL) {
#ifdef DEBUG
            if (!KeyPad[i].str[0]) {
		fprintf(debug_file, "KeyPad %d is empty\n",i);
	    } else
            fprintf(debug_file, "Checking %d '%s == %s'\n", i, KeyPad[i].str, chbuf);
	    if (strcmp(KeyPad[i].str, chbuf) == 0) {
              if (prefixed) 
                fprintf(debug_file, "Got match on %s, keycode %d\n"
                        , chbuf, KeyPad[i].keycode );
              else
                fprintf(debug_file, "Got match on ESC-%s, keycode %d\n"
                        , chbuf+1, KeyPad[i].keycode );                            
              return(KeyPad[i].keycode);
            }
	    if ((prefixed) && (strncmp(KeyPad[i].str, chbuf, endbuf) == 0)) {
              fprintf(debug_file, "Got partial match on ESC-%s with keypad%d\n"
                      , chbuf+1, i);              
              ThereisHope = TRUE;
            }            
#else
	    if (strcmp(KeyPad[i].str, chbuf) == 0)
		return(KeyPad[i].keycode);
            if ((prefixed) && (strncmp(KeyPad[i].str, chbuf, endbuf) == 0))
		ThereisHope = TRUE;
#endif
	    i++;
	}
	if (!ThereisHope) break;
#ifdef CHECK_BYTES_AVAILABLE
	wait_keyboard_io(FALSE); /* check descriptors */
	if (bytes_available() <= 0) break; /* should have been a burst! */
#endif /* CHECK_BYTES_AVAILABLE */
	chbuf[endbuf++] = GetKey() & 0x7f;
	chbuf[endbuf]   = '\0';
    }

    return(chin);
}

void SCR_SetStringAttribute(attr_string, nlines)
char *attr_string;
int nlines;
{
  if (attr_string != NIL)
    tputs(attr_string, nlines, _putchar_rtn);
}

int SCR_AlternateCharSet(int on_off)
{
  void SCR_SetBlink();
  if (tc_start_alt_set == NIL) return(FALSE);
  if (on_off == TRUE && cur_mode == GRAFIC) return(TRUE);
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)  || 1
  if (on_off == FALSE) SCR_UpdateScreen();
  ATTRSET((on_off == TRUE) ? A_ALTCHARSET : A_NORMAL);
#else
  SCR_UpdateScreen();
  if (on_off == TRUE) {
    if (cur_mode == HILITE || cur_mode == HILITC) SCR_SetHighlight(FALSE);
    else if (cur_mode == BLINKH) SCR_SetBlink(FALSE);
  }
  SCR_SetStringAttribute((on_off ? tc_start_alt_set : tc_end_alt_set), 1);
#endif
  cur_mode = (on_off ? GRAFIC : NORMAL);
  return(TRUE);
}

void SCR_SetBlink(on_off)
int on_off;
{
  if (tc_set_blink == NIL) return;
  if (on_off == TRUE && (cur_mode == BLINKH)) return;
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
  if (on_off == FALSE) SCR_UpdateScreen();
  ATTRSET((on_off == TRUE) ? A_BLINK : A_NORMAL);
#else
  SCR_UpdateScreen();
  if (on_off == TRUE) {
    if (cur_mode == GRAFIC) SCR_AlternateCharSet(FALSE);
    else if (cur_mode == BLINKH) SCR_SetBlink(FALSE);
  }
  SCR_SetStringAttribute((on_off ? tc_set_blink : tc_set_all_off), 1);
#endif
  cur_mode = (on_off ? BLINKH : NORMAL);
}

void reset_cur_mode(void)
{
  if (cur_mode == GRAFIC) SCR_AlternateCharSet(FALSE);
  else if (cur_mode == HILITE || cur_mode == HILITC) SCR_SetHighlight(FALSE);
  else if (cur_mode == BLINKH) SCR_SetBlink(FALSE);
  if (!nostandout) standend();
  cur_mode = NORMAL;
}

void SCR_HomeCursor (int row)
{
    move(row, 0);
}

void SCR_ClearScreen(void)
{
  clear();

  memset(refresh_buffer, SPC_KEY, screen_rows * screen_cols);
  memset(refresh_attr,   NORMAL,  screen_rows * screen_cols);
}

void SCR_NoRawState(void)
{
  SCR_SetStringAttribute(tc_cap_ke, 1);
  reset_cur_mode();
  noraw();
  echo();
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
  nocbreak();
  nl();
#endif
  cbreak();
  SCR_ClearScreen();
  SCR_HomeCursor(screen_rows-1);
  SCR_UpdateScreen();
}

static void vt100_hook(void)
{
  if (cur_char_set == VT_C_SET)
    _putchar_rtn(KEYCODE_CTRL_O); /* reset the vt100 alt char set */
}

void SCR_RedrawScreen(void)
{
  vt100_hook();
  SCR_ClearScreen();
  wclear(curscr);
  WIN_RedrawScreen();
}

void SCR_RawState(void)
{
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
  nonl();
  cbreak();
#endif
  noecho();
  raw();
  vt100_hook();
  reset_cur_mode();
  SCR_SetStringAttribute(tc_cap_ks, 1);
  SCR_RedrawScreen();
}

void SuspendProcess()
{
#ifdef SIGTSTP
  if (!gCapturedShell) {
    checkpoint_openfiles();
    SCR_NoRawState();
    endwin();
    printf("\nUse 'fg' to return.\n");
    fflush(stdout);
    kill(getpid(),SIGTSTP);
    SCR_RawState();
  } else
    SCR_Beep();
#else
  SCR_Beep();
#endif /* SIGTSTP */
}

void SCR_SetCursorPosn(row, col)
int row, col;
{
  if (row >= screen_rows || col >= screen_cols)
    move (0, 0);
  else {
    if (!CursorSaved) {
      cursor_row = row;
      cursor_col = col;
    }
    move(row, col);
  }
}

int SCR_CursorOff (void) 
{ 
  if (CursorSaved) return FALSE;
  saved_row = cursor_row;
  saved_col = cursor_col;
  cursor_row = screen_rows;
  cursor_col = 0;
  CursorSaved = TRUE;
  move(cursor_row, cursor_col);
  return(TRUE);
}

void SCR_CursorOn (void)
{
  if (CursorSaved && cursor_row == screen_rows && cursor_col == 0) {
    cursor_row = saved_row;
    cursor_col = saved_col;
  }
  move(cursor_row, cursor_col);
  CursorSaved = FALSE;
}

void SCR_ClearLine (int line_no)
{
    int y, x;

    getyx(stdscr, y, x);
    move(line_no, 0);
    clrtoeol();
    move(y, x);
}

char *SCR_GetCap(char *a_cap,char *buf)
{
  char *TCbufptr, *attr_string;
  TCbufptr = &TCbuf[0];

  attr_string = tgetstr(a_cap, &TCbufptr);

  if (buf && attr_string) {
    strncpy(buf, attr_string, MAX_TC_LEN);
    buf[MAX_TC_LEN] = 0;
    attr_string = buf;
  }
  return(attr_string);
}

void SCR_SendIS(void)
{
  SCR_SetStringAttribute(SCR_GetCap(init_string_cap, NIL), 1);
  SCR_SetStringAttribute(tc_cap_ks, 1); /* into keypad xmit */
#ifdef ultrix
  move(0,0);
  if (HO) tputs(HO, 0, _putchar_rtn);
#endif
}

void SCR_SetupVideo(void)
{

    void SCR_UseCharSet();
    int  i;

#ifdef DEBUG
    debug_file=fopen("/tmp/ezmail-debug", "a+");
    if (!debug_file)
      debug_file=fopen("/dev/null", "a+");
    setlinebuf(debug_file);
#endif
    curswin=initscr();
    if (CM == NIL) {
	printf("TERM is undefined.\n");
	printf("Your terminal lacks the ability to run this program.\n");
	exit(1);
    }
    tc_cap_ks = SCR_GetCap(enter_keypad_cap, CAP_ks);
    tc_cap_ke = SCR_GetCap(exit_keypad_cap, CAP_ke);
    SCR_SetStringAttribute(tc_cap_ks, 1); /* into keypad xmit */

    for (i=0; KeyPad[i].str != NIL; i++) {
      char *tmpptr;
      KeyPad[i].str[0] = '\0';
      if ((tmpptr = SCR_GetCap(KeyPad[i].termcap_id, NIL)) != NIL) {
	if ((*tmpptr) && (*tmpptr == ESC_KEY)) { /* starts with ESC key */
#ifdef DEBUG
          fprintf(debug_file,"got esc-sequence for (%d) %s\nseq is ESC-'%s'\n", i,
                  KeyPad[i].termcap_id, tmpptr+1);
#endif
	  strncpy(KeyPad[i].str, tmpptr, MAX_TC_LEN);
	  KeyPad[i].str[MAX_TC_LEN] = '\0';
	} 
#ifdef DEBUG
	else { fprintf(debug_file, "(%d) %s is non-esc sequense '%s'\n", i, KeyPad[i].termcap_id, tmpptr); }
#endif
      }
#ifdef DEBUG
	else { fprintf(debug_file, "(%d) %s isn't defined\n", i, KeyPad[i].termcap_id); }
#endif

    }
    
    nostandout = (SO == NIL);
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
    nonl(); 
    cbreak(); 
#endif
    noecho(); 
    raw();

    screen_cols = COLS;
    screen_rows = LINES;
    status_row  = screen_rows - 1;

    AttrTablePtr = Termcap_AttrTable;

    tc_start_alt_set = SCR_GetCap(enter_alt_charset_cap, CAP_as);
    tc_end_alt_set   = SCR_GetCap(exit_alt_charset_cap, CAP_ae);
    tc_set_bold      = SCR_GetCap(enter_bold_cap, CAP_md);
    tc_set_blink     = SCR_GetCap(enter_blink_cap, CAP_mb);
    tc_set_all_off   = SCR_GetCap(exit_attribute_cap, CAP_me);
    if (tc_start_alt_set && tc_end_alt_set) {
      char *sp, name[100];
#if defined(ultrix) || (defined(sun) && !defined(__SVR4)) 
      char bigbuf[1024];
#endif
      
      name[0]='\0';

#if defined(ultrix) || (defined(sun) && !defined(__SVR4))      
    if ((sp = getenv("TERM")) && (tgetent(&bigbuf[0], sp) != -1))
      longname(&bigbuf[0], &name[0]);
#else
      sp=longname();
      if (sp)
          strcpy(name,sp);
#endif

      if ((strncmp(&name[0], "vt100", 5) == 0) ||
          (strncmp(&name[0], "vt102", 5) == 0) ||
          (strncmp(&name[0], "vs100", 5) == 0) ||
          (strncmp(&name[0], "linux", 5) == 0) ||
          (strncmp(&name[0], "xterms", 6) == 0) ||
          (strncmp(&name[0], "xterm", 5) == 0) ||
          (strncmp(&name[0], "vt200", 5) == 0) ||
	  (strncmp(&name[0], "X11 terminal emulator", 3) == 0) || 
          (strncmp(&name[0], "vt220", 5) == 0)) SCR_UseCharSet(VT_C_SET);
      else if (strcmp(&name[0], "h19") == 0) SCR_UseCharSet(ZT_C_SET);
      else SCR_UseCharSet(NO_C_SET);
    } else {
      SCR_UseCharSet(NO_C_SET);
      sleep(2);
    }
    
    refresh_buffer = (unsigned char *)malloc(screen_rows * screen_cols);
    refresh_attr   = (unsigned char *)malloc(screen_rows * screen_cols);

    vt100_hook();
    SCR_ClearScreen();
    SCR_UpdateScreen();
    init_signals();
}

void SCR_NoHighlight()
{
  tc_set_bold = NIL;
}

void SCR_InitCharMap(hl, vl, ul, ll, ur, lr, lb, rb)
char hl, vl, ul, ll, ur, lr, lb, rb;
{
  int i;

  for (i=0; i<128; i++) charmap[i] = i;
  for (i=128; i<256; i++) charmap[i] = '$';

  charmap[0315] = hl;
  charmap[0272] = vl;
  charmap[0311] = ul;
  charmap[0310] = ll;
  charmap[0273] = ur;
  charmap[0274] = lr;
  charmap[0265] = lb;
  charmap[0306] = rb;
}

void SCR_UseCharSet(int num)
{
  tc_start_alt_set = (*CAP_as ? CAP_as : NIL);
  tc_end_alt_set   = (*CAP_ae ? CAP_ae : NIL);

  if (tc_start_alt_set && tc_end_alt_set) {
    if (num == VT_C_SET)
      SCR_InitCharMap(VT_HL, VT_VL, VT_UL, VT_LL,
		      VT_UR, VT_LR, VT_PL, VT_PR);
    else if (num == ZT_C_SET)
      SCR_InitCharMap(H19_HL, H19_VL, H19_UL, H19_LL,
		      H19_UR, H19_LR, H19_PL, H19_PR);
    else
      num = NO_C_SET;
  }
  if (num == NO_C_SET) {
    SCR_InitCharMap('-', '|', '+', '+', '+', '+', '|', '|');
    tc_start_alt_set = NIL;
    tc_end_alt_set   = NIL;
  }
  cur_char_set = num;
}



void SCR_RestoreVideoEnv (void)
{
  SCR_NoRawState();
  endwin();
  erase();
}

int SCR_CanHighlight(void)
{
  return(tc_set_bold ? TRUE : FALSE);
}

void SCR_SetHighlight(on_off)
int on_off;
{
  if (tc_set_bold == NIL) return;
  if (on_off == TRUE && (cur_mode == HILITE || cur_mode == HILITC)) return;
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
  if (on_off == FALSE) SCR_UpdateScreen();
  ATTRSET((on_off == TRUE) ? A_BOLD : A_NORMAL);
#else
  SCR_UpdateScreen();
  if (on_off == TRUE) {
    if (cur_mode == GRAFIC) SCR_AlternateCharSet(FALSE);
    else if (cur_mode == BLINKH) SCR_SetBlink(FALSE);
  }
  SCR_SetStringAttribute((on_off ? tc_set_bold : tc_set_all_off), 1);
#endif
  cur_mode = (on_off ? HILITE : NORMAL);
}

static int g_background = Background(BLACK);

void SCR_InitBackground(void)
{
  g_background = BLACK; /*GetLpfColor(SCREEN_BACKGROUND);*/
}

int SCR_GetAndSetBackground(int attr)
{
  int s = g_background;
  g_background = attr;
  return(s);
}

void SCR_Write(unsigned char *data,int row, int col, int len, int attr)
{
  if (row >= screen_rows || col >= screen_cols) 
    return;

  if (len > 0) {
    int shift=0;
    register unsigned char ch;

    if (data || attr == RVIDEO || attr == RINVIS) {
	if (attr != cur_mode) {
	    switch (cur_mode) {
	    case GRAFIC:
		SCR_AlternateCharSet(FALSE);
		break;
	    case HILITE:
	    case HILITC:
		SCR_SetHighlight(FALSE);
		break;
	    case BLINKH:
		SCR_SetBlink(FALSE);
		break;
	    }
	    if (attr == HILITE || attr == HILITC) SCR_SetHighlight(TRUE);
	    else if (attr == GRAFIC) SCR_AlternateCharSet(TRUE);
	    else if (attr == BLINKH) SCR_SetBlink(TRUE);
	    cur_mode = attr;
	}
    }

    if (*(AttrTablePtr + attr)) {
      standout();
      shift = nostandout;
    }

    ATTRSET(A_NORMAL);
    switch (attr) {
    case RVIDEO:
    case RINVIS:
    case WHTBLU:
      if (!nostandout)
	ATTRSET(A_STANDOUT);
      break;
    case GRAFIC:
      if (tc_start_alt_set)
	ATTRSET(A_ALTCHARSET);
      break;
    case HILITE:
    case HILITC:
      if (tc_set_bold)
	ATTRSET(A_BOLD);
      break;
    case BLINKH:
      if (tc_set_blink)
	ATTRSET(A_BLINK);
      break;
    }

    move(row, col);

    while (len-- > 0) {
      int idx = row * screen_cols + col;
      if (col >= screen_cols) break;
      ch = ( data && *data) ? charmap[*(data++)] : SPC_KEY;
      if (ch != SPC_KEY && (attr == RINVIS || attr == INVISI)) ch = '*';
      else if (ch < SPC_KEY) {
	if (ch == TAB_KEY) ch = SPC_KEY;
	else {
          ch = ' ';
        }
      } else {
	if (shift && isascii(ch) && islower(ch)) ch = toupper(ch);
      }
      /*#if !defined(__hpux) && !defined(__SVR4) && !defined(__linux__)
      if (ch != SPC_KEY) {
	if (refresh_buffer[idx] == ch && refresh_attr[idx] != attr) {
	  stdscr->_y[row][col] = SPC_KEY;
	  curscr->_y[row][col] = SPC_KEY;
	}
      }
#endif */
      refresh_buffer[idx] = ch;
      refresh_attr[idx] = attr;
      addch(ch);
      col++;
    }
    if (*(AttrTablePtr + attr))
      standend();
#if defined(__hpux) || defined(__SVR4) || defined(__linux__)
    ATTRSET(A_NORMAL);
#endif
    move(cursor_row, cursor_col);
  }
}

void SCR_Line(int rows, int cols, int ypos, int xpos, int attr, write_proc_t write_proc)
{
    int  i;
    char strbuf[STR255LEN];

    if (write_proc == NULL) write_proc = SCR_Write;

    if (SCR_AlternateCharSet(TRUE)) attr = GRAFIC;

    memset(strbuf, (int)ACS_HLINE/*'\315'*/, cols);

    for (i=ypos; i<ypos+rows; i++)
      (*write_proc)(strbuf, i, xpos, cols, attr);
}

void SCR_RightT(int r,int c,int attr,write_proc_t write_proc)
{
    if (write_proc == NULL) write_proc = SCR_Write;

    if (SCR_AlternateCharSet(TRUE)) attr = GRAFIC;

    (*write_proc)("\306", r, c, 1, attr);
}

void SCR_LeftT(int r,int c,int attr,write_proc_t write_proc)
{
    if (write_proc == NULL) write_proc = SCR_Write;

    if (SCR_AlternateCharSet(TRUE)) attr = GRAFIC;

    (*write_proc)("\265", r, c, 1, attr);
}

void clear_box(int rows, int cols, int ypos, int xpos,write_proc_t write_proc)
{
    int  i;

    if (write_proc == NULL) write_proc = SCR_Write;

    (*write_proc)(NIL, ypos, xpos, cols, NORMAL);
    for (i=1; i<rows-1; i++) {
      (*write_proc)(NIL,i+ypos, xpos, 1, NORMAL);
      (*write_proc)(NIL,i+ypos, xpos+cols-1, 1, NORMAL);
    }

    (*write_proc)(NIL, ypos+rows-1, xpos, cols, NORMAL);
}

void SCR_Box(int rows, int cols, int ypos, int xpos, int attr,
	     write_proc_t write_proc)
{
    int  i;
    char strbuf[STR255LEN];

    if (write_proc == NULL) write_proc = SCR_Write;

    if (SCR_AlternateCharSet(TRUE)) attr = GRAFIC;
    if (attr == RINVIS && nostandout) attr = NORMAL;

    /*    if (attr != GRAFIC && type == DOUBLE_LINE_BOX)
      memset(strbuf, (int)'=', cols);
    else */
      memset(strbuf, (int)'\315', cols);

    strbuf[0] = '\311'; strbuf[cols-1] = '\273';
    (*write_proc)(strbuf, ypos, xpos, cols, attr);
    for (i=1; i<rows-1; i++) {
      (*write_proc)("\272",i+ypos, xpos, 1, attr);
      (*write_proc)("\272",i+ypos, xpos+cols-1, 1, attr);
    }
    strbuf[0] = '\310';
    strbuf[cols-1] = '\274';
    (*write_proc)(strbuf, ypos+rows-1, xpos, cols, attr);
}

void SCR_LineWTitle(rows, cols, ypos, xpos, attr, write_proc, type, title)
int rows, cols, ypos, xpos, attr;
write_proc_t write_proc;
char *title;
{
  if (!title || !*title)
      SCR_Line(rows, cols-2, 0, 1, attr, write_proc);
  else {
    int tlen, len;
    len = imin(strlen(title), cols-MIN_WIDTH);
    tlen = (cols - len) / 2;
    SCR_Line(1, tlen, 0, 1, attr, write_proc);
    SCR_LeftT(0, tlen, attr, write_proc);
    SCR_RightT(0, tlen+len+1, attr, write_proc);
    SCR_Line(1, cols-(tlen+len+3), 0,
	     tlen+len+2, attr, write_proc);
  }
}

void SCR_BoxWTitle(ypos, xpos, rows, cols, attr, write_proc, type, title)
int rows, cols, ypos, xpos, attr;
write_proc_t write_proc;
char *title;
{
    int  i;
    char strbuf[STR255LEN];

    if (write_proc == NULL) write_proc = SCR_Write;

    if (SCR_AlternateCharSet(TRUE)) attr = GRAFIC;
    if (attr == RINVIS && nostandout) attr = NORMAL;

    if (attr != GRAFIC && type == DOUBLE_LINE_BOX)
      memset(strbuf, (int)'=', cols);
    else
      memset(strbuf, (int)'\315', cols);

    strbuf[0] = '\311'; strbuf[cols-1] = '\273';

    (*write_proc)(&strbuf[0], ypos, xpos, 1, attr);
    (*write_proc)(&strbuf[cols-1], ypos, xpos+cols-1, 1, attr);

    SCR_LineWTitle(1, cols, ypos, xpos, attr, write_proc, type, title);

    for (i=1; i<rows-1; i++) {
      (*write_proc)("\272",i+ypos, xpos, 1, attr);
      (*write_proc)("\272",i+ypos, xpos+cols-1, 1, attr);
    }
    strbuf[0] = '\310';
    strbuf[cols-1] = '\274';
    (*write_proc)(strbuf, ypos+rows-1, xpos, cols, attr);
}

static Boolean status_line_is_clear = TRUE;

void SCR_WriteStatusLine(char *str)
{
  SCR_Write(str, status_row, 0, screen_cols-1, NORMAL);
  SCR_UpdateScreen();
  status_line_is_clear = FALSE;
}

void SCR_ClearStatusLine(void)
{
/*  if (!status_line_is_clear) */
    SCR_WriteStatusLine(NIL);
  status_line_is_clear = TRUE;
}

int SCR_isStatusLineClear(void)
{
  return(status_line_is_clear);
}

void SCR_WriteTextOnButtonLine(text)
char *text;
{
  SCR_WriteStatusLine(text);
}

void SCR_WriteWithAttr(text, r, c, len, attr)
char *text;
int r, c, len;
char *attr;
{
  if (text && attr) {
    int i;
    for (i=c; i<c+len; i++) SCR_Write(text++, r, i, 1, *attr++);
  }
}

int SCR_UpCaseRepresentation(int key_code)
{
  int new_key;
  if (key_code >= 256) {
    new_key = key_code - 256;
    if (islower(new_key)) new_key = toupper(new_key);
    new_key += 256;
  } else {
    if (islower(key_code)) new_key = toupper(key_code);
    else new_key = key_code;
  }
  return(new_key);
}

char *SCR_GetKeyText(int key_code)
{
  static char key_buf[10];

  key_buf[0] = 0;
  if (key_code >= 256) {
    strcpy(key_buf, "Esc");
    key_code -= 256;
  } else if (key_code <= 26 && key_code!=TAB_KEY && key_code!=KEYCODE_RETURN) {
    strcpy(key_buf, "Ctl");
    key_code += 'A' - 1;
  }
  if (key_code == TAB_KEY)
    strcat(key_buf, "TAB");
  else if (key_code == KEYCODE_RETURN)
    strcpy(key_buf, "RETURN");
  else {
    int len = strlen(key_buf);
    if (len == 0) {
      key_buf[len++] = '\"';
      key_buf[len++] = (char)key_code;
      key_buf[len++] = '\"';
    } else
      key_buf[len++] = (char)key_code;
    key_buf[len] = '\0';
  }
  return(key_buf);
}

/*
 * Some of these routines are taken from file cr_put.c in the curses library.
 * These routines match closely to their curses counter-parts.  In order for
 * these routines to work, you have to have the curses library.
 */

static int cur_y, cur_x;
static int new_y, new_x;
static int num_outs;

void count_outs(int c)
{
  num_outs--;
}

int move_cursor(int (*out_proc)())
{
#define HARDTABS 8

  WINDOW *_win;
  int i, j, k;

  if (HO) {
    if (GT) i = (new_x / HARDTABS) + (new_x % HARDTABS);
    else    i = new_x;
    if (new_x >= cur_x) { /* to the right */
      j = new_x / HARDTABS - cur_x / HARDTABS;
      if (GT && j) j += new_x % HARDTABS;
      else j = new_x - cur_x;
    } else { /* to the left */
      if (cur_x - new_x <= i && (BS || BC)) i = j = cur_x - new_x;
      else j = i + 1;
    }
    k = iabs(cur_y - new_y);
    j += k;
    if (i + new_y < j || (!UP && new_y < cur_y)) {
      tputs(HO, 0, out_proc);
      cur_x = cur_y = 0;
    } else if (LL) {
      k = (LINES - 1) - new_y;
      if (i + k + 2 < j && (k <= 0 || UP)) {
	tputs(LL, 0, out_proc);
	cur_x = 0; cur_y = LINES - 1;
      }
    }
  } else if (!UP && new_y < cur_y) return(-1);

  if (GT) i = new_x % HARDTABS + new_x / HARDTABS;
  else    i = new_x;
  j = cur_x - new_x;

#ifdef NONL
  if ((NONL || !_pfast_thingy) && cur_y < new_y) goto bypass_cr;
#endif

  if (NC && cur_y >= new_y) goto bypass_cr;

  if ((j > i + 1) || (cur_x > cur_y && !BS && !BC)) {
    if (CR) tputs(CR, 0, out_proc);
    else (*out_proc)('\r');
    if (NC) {
      if (NL) tputs(NL, 0, out_proc);
      else (*out_proc)('\n');
      cur_y++;
    }
    cur_x = 0;
  }

bypass_cr:
  while(cur_y < new_y) {
    cur_y++;
    if (NL) tputs(NL, 0, out_proc);
    else (*out_proc)('\n');
    if (num_outs < 0) return(num_outs);
#ifdef NONL
    if (NONL || _pfast_thingy == 0) cur_x = 0;
#endif
  }

  if (BT) k = strlen(BT);

  while (cur_x > new_x) {
    if (num_outs < 0) return(num_outs);
    cur_x--;
    if (BC) tputs(BC, 0, out_proc);
    else (*out_proc)('\b');
  }

  while (cur_y > new_y) {
    cur_y--;
    tputs(UP, 0, out_proc);
    if (num_outs < 0) return(num_outs);
  }
  if (GT && new_x - cur_x > 1) {
    while(TRUE) {
      i = tabcol(cur_x, HARDTABS);
      if (i > new_x) break;
      if (TA) tputs(TA, 0, out_proc);
      else (*out_proc)('\t');
      cur_x = i;
    }
    if (new_x - cur_x > 4 && i < COLS && (BC || BS)) {
      if (TA) tputs(TA, 0, out_proc);
      else (*out_proc)('\t');
      cur_x = i;
      while (cur_x > new_x) {
	cur_x--;
	if (BC) tputs(BC, 0, out_proc);
	else (*out_proc)('\b');
      }
    }
  }

  while (cur_x < new_x) {
    if (_win != NULL) {
      register int idx = cur_y * screen_cols + cur_x;
      if (refresh_buffer[idx]!=SPC_KEY && (tc_start_alt_set || tc_set_bold)) {
	if (refresh_attr[idx] != cur_mode) goto use_nd;
      }
      /*#if !defined(__hpux) && !defined(__SVR4) && !defined(__linux__)
      i = curscr->_y[cur_y][cur_x];
      if ((i & _STANDOUT) == (curscr->_flags & _STANDOUT)) 
	(*out_proc)(i);
      else 
#endif */
	goto use_nd;
    } else 
use_nd:
      if (ND) tputs(ND, 0, out_proc);
      else (*out_proc)(' ');
    cur_x++;
    if (num_outs < 0) break;
  }

  return(num_outs);
}

int tabcol(int col, int ts)
{
  int offset;
  if (col >= COLS) {
    offset = COLS * (col / COLS);
    col -= offset;
  } else
    offset = 0;

  return(col + ts - (col % ts) + offset);
}

void tstp(int sig)
{
  /* override curses signal handling for TSTP.  If we are stopped
     due to SIGTSTP in an editor, we don't want curses do refresh
     all over the editor's screen when we come back.
  */

#if defined(__SVR4) || defined(__linux__)
  struct sigaction sa;
  signal(SIGTSTP, SIG_DFL);
  kill(0, SIGTSTP);
  sa.sa_handler=tstp;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags=SA_NODEFER;
  sigaction(SIGTSTP, &sa, NULL);
#else
#ifdef SIGTSTP
  int omask;
  omask = sigsetmask(sigblock(0) & ~(sigmask(SIGTSTP)));
  signal(SIGTSTP, SIG_DFL);
  kill(0, SIGTSTP);
/*  sigblock(sigmask(SIGTSTP)); */
  signal(SIGTSTP, tstp);
#endif /* SIGTSTP */
#endif
}

void sig_handler(int sig)
{
  save_the_world();
  SCR_NoRawState();
  endwin();
  signal(sig, SIG_DFL);
  kill(getpid(), sig);
}

void winch_handler(sig, code, scp)
int sig, code;
#ifdef __SVR4
ucontext_t *scp;
#else
struct sigcontext *scp;
#endif
{
  unsigned char *nbuff;
  unsigned char *nattr;
#ifdef __hpux
  _winchTrap(sig, code, scp); /* call curses winch-handler routine */
  signal(SIGTSTP, tstp);      /* trap these guys again */
  signal(SIGWINCH, winch_handler);
#else
  struct winsize ws;
  if (ioctl(0, TIOCGWINSZ, &ws) < 0) return;
  SCR_NoRawState();
  COLS = ws.ws_col;
  LINES = ws.ws_row;

  echo();
/*
  nocbreak();
  nl();
*/
  noraw();

  initscr();

/*
  nonl();
  cbreak();
*/
  noecho();
  raw();
#endif
  screen_cols = COLS;
  screen_rows = LINES;
  status_row  = screen_rows - 1;
  if ((nbuff = (unsigned char *)malloc(screen_rows * screen_cols))==0)
    return;
  if ((nattr = (unsigned char *)malloc(screen_rows * screen_cols))==0)
    return;
  free(refresh_buffer);
  free(refresh_attr);
  refresh_buffer = nbuff;
  refresh_attr = nattr;
  SCR_ClearScreen();
  WIN_Reinitialize();
  SCR_RedrawScreen();
  /* JCS - when we resize, we have to turn the keypad back on for 
   * some reason
   */
  SCR_SetStringAttribute(tc_cap_ks, 1);
  SCR_ClearKeyboard();
}

#define ITIMER_INTERVAL (60)  /* 60 second interval timer */

static int itimer_value = 0;

int get_itimer_value(void)
{
  return(itimer_value);
}

void set_itimer_value(int val)
{
  itimer_value = val;
}

void sigalm_handler(sig, code, scp)
int sig;
{
  itimer_value += ITIMER_INTERVAL;
#ifdef __hpux
  signal(SIGALRM, sigalm_handler);
#endif
}

void init_signals(void)
{
#if defined(__SVR4) || defined(__linux__)
struct sigaction sa;
#endif
  signal(SIGINT, SIG_IGN);
  signal(SIGILL, sig_handler);
  signal(SIGIOT, sig_handler);
#ifdef SIGEMT
  signal(SIGEMT, sig_handler);
#endif
#ifdef SIGBUS
  signal(SIGBUS, sig_handler);
#endif
#ifdef SIGSYS
  signal(SIGSYS, sig_handler);
#endif
  signal(SIGHUP, sig_handler);
  signal(SIGSEGV, sig_handler);
#if defined(__SVR4) || defined(__linux__)
  sa.sa_handler=tstp;
  sigemptyset(&sa.sa_mask);
  sa.sa_flags=SA_NODEFER;
  sigaction(SIGTSTP, &sa, NULL);
  sa.sa_flags=0;
  sa.sa_handler=winch_handler;
  sigaction(SIGWINCH, &sa, NULL);
  sa.sa_handler=sigalm_handler;
  sigaction(SIGALRM, &sa, NULL);
#else
  signal(SIGTSTP, tstp);
  signal(SIGWINCH, winch_handler);
  signal(SIGALRM, sigalm_handler);
#endif
  {
    /* interval timer set to ITIMER_INTERVAL defined in common.h */
    struct itimerval value;
    value.it_value.tv_sec  = value.it_interval.tv_sec  = ITIMER_INTERVAL;
    value.it_value.tv_usec = value.it_interval.tv_usec = 0;
    setitimer(ITIMER_REAL, &value, (struct itimerval *)0);
  }
}

int get_button_index(char *text,int size, int x, int y)
{
  return(-1);
}

int SCR_iscolor(void)
{
  return(FALSE);
}

int need_attention(char *msg)
{
  if (SCR_Keyboard_io()) {
    if (SCR_ClearKeyboard()) {
      if (DLG_GetBooleanFromUser(msg, FALSE)) return(TRUE);
      SCR_Beep();
      SCR_WriteStatusLine(MSG_Attention);
    }
  }
  return(FALSE);
}
