/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */



#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <libgen.h>

#include "common.h"
#include "comkeys.h"
#include "menus.h"
#include "screen.h"
#include "list.h"
#include "and_keys.h"
#include "and_scr.h"
#include "dialog.h"
#include "console.h"
#include "windows.h"
#include "text.h"
#include "document.h"
#include "clinet.h"
#include "and_mach.h"
#include "ezconfig.h"
#include "rli.h"

int handle_save_attachment()
{
  char 
    filename[STR255LEN],
    attachmentNumber[STR255LEN];

  filename[0] = '\0';
  attachmentNumber[0] = '\0';
  
  /* Query the user as to which attachment to save and to where
   * it should be saved.
   */
  if ( !DLG_GetAttachmentInfo ( filename,
				attachmentNumber ) )
    return (FALSE);

  { 
    int 
      pos = 0,
      lastPos = 0,
      fd;

    char
      * attachment;
    
    fd = open ( filename, O_CREAT | O_WRONLY, S_IRUSR | S_IWUSR );
    
    if ( fd < 0 ) { 
      CON_Write ( "Cannot open filename: " );
      CON_WriteLine ( (char*)filename );
      return ( FALSE );
    }

    if ( attachmentNumber == NULL || 
         strlen ( attachmentNumber ) == 0 ) 
      attachment = "TEXT";
    else
      attachment = (char*)attachmentNumber; 
    
    do { 

      lastPos = pos;

      CON_ShowTXStatus ( pos / 1024 );
      pos +=  do_getmsg_to_buffer ( do_getcurrentg(), 
                                    GET_LIST_SELECTION ( CAPTION_RL ),
                                    attachment,
                                    pos );
      
      /* If we're downloading a large message, you might want the user
       * to cancel the operation in the middle.
       */
      if (SCR_Keyboard_io() && 
          SCR_ClearKeyboard() &&
          DLG_GetBooleanFromUser(MSG_ServerWait, FALSE)) {
        break;
      }
      
      if ( pos <= 0 || pos == lastPos || g_msgbuffer == NULL ) 
        break;
      
      write ( fd, g_msgbuffer, g_msg_decoded );

    } while ( 1 );

    close ( fd );
  }

  CON_WriteLine ( "Done." );
  
  return (TRUE);
}

/* Save the entire email body to the given filename */
int handle_save_all_to_file ( const char * filename )
{
  int
    fd;

  fd = open ( filename, O_CREAT | O_WRONLY );
  
  if ( fd < 0 ) { 
    CON_Write ( "Cannot open filename: " );
    CON_WriteLine ( (char*)filename );
    return ( FALSE );
  }

  do_save_to_file ( fd, NULL );

  close ( fd );

  return ( TRUE );
}
