/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */



#include <config.h>

#include <sasl.h>
#include <saslutil.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>
#include <errno.h>
#include <netinet/in.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <ctype.h>


#include "mmspfio.h"
#include "rli.h"
#include "rlci.h"
#include "list.h"
#include "password.h"

#include "dialog.h"
#include "clinet.h"
#include "text.h"

#include "imclient.h"
#include "imparse.h"

#include "skip-list.h"

#include "and_scr.h"
#include "console.h"
#include "windows.h"

#include "xmalloc.h"

#include "and_file.h"
#include "mime.h"

//#define IMAP_PROTOCOL_DEBUG

#ifdef IMAP_PROTOCOL_DEBUG
#undef imclient_send
#define imclient_send(CLIENT,PROC,ROCK,...) do { \
   fprintf ( stderr, "****** @%s:%d SEND ******\n", __FILE__, __LINE__ ); \
   fprintf ( stderr, __VA_ARGS__ ); \
   fprintf ( stderr, "\n" ); \
   imclient_send ( CLIENT, PROC, ROCK, __VA_ARGS__ ); \
    } while ( 0 )

#define IMAP_RESPONSE(REPLY) do { \
  fprintf ( stderr, "***** @%s:%d RECEIVE *****\n", __FILE__, __LINE__ ); \
  fprintf ( stderr, "keyword: %s, Msgno: %ld\nText:\n%s\n", REPLY->keyword, REPLY->msgno, REPLY->text ); \
  } while ( 0 );
#else
#define IMAP_RESPONSE(REPLY) 
#endif

#define TRANSFER_BLOCK_SIZE 4096

extern void rl_parerr();

extern char *index();

extern char **gl_envp;

int msg_read;

#ifndef DEFAULT_MMS_SERVER
#define DEFAULT_MMS_SERVER "/usr/local/bin/mms"
#define LOCAL_MMS_SERVER   "/etc/mms"
#endif /* DEFAULT_MMS_SERVER */

#define GUARDIAN_PORT 900
#define DEFAULT_SERVICE "ezmail"

long recieved_bytes=0;
void (*rl_incomming_func)();


#define C_DATA (0)		/*packet data connection*/
#define C_CONS (1)		/*console connection*/
#define C_NUMO (2)		/*number of connections*/

int never_sel();
int clie_cons_enable();

#define NOTFINISHED 0
#define IMAP_OK 1
#define IMAP_NO 2
#define IMAP_BAD 3
#define IMAP_EOF 4

/* Used in all the functions in case the connection goes dead */
#define RESTART_SPOT     \
if (0) {    \
 restart: \
    trys++; \
    if (trys > 2) return -1; \
    CON_WriteLine("Connection closed. Reconnecting..."); \
    mn_clie_connect(g); \
} 

typedef struct changed_item_s {
    char *name;
    int unseen;
} changed_item_t;

typedef struct set_s {
    int low;
    int high;
} set_t;

int g_rot13_mode = 0;
int g_fullheaders_mode = 0;

static void rot13_line(char *line);


struct globals_s {
    char *server_hostname;
    char *userid;

    struct imclient *imclient_conn;    
    int cmd_done;
    int cmds_finished;

    char *selected_mbox;
    int   deletedMessages;

    int exists;
    int unseen;

    int inboxunseen;

    int quotaused;
    int quotalimit;
};

#define MAX_GLOBS 5

globals_t *globs[MAX_GLOBS];
int num_globs = 0;
globals_t *currentmbox_g = NULL;

static int mn_clie_connect(globals_t *g);

int spot = 0;

globals_t *do_getcurrentg(void)
{
    if (currentmbox_g) return currentmbox_g;

    /* Otherwise just return the first g */
    if (num_globs == 0) return NULL;
    return globs[0];
}

/***********************
 * Parse a mech list of the form: ... AUTH=foo AUTH=bar ...
 *
 * Return: string with mechs seperated by spaces
 *
 ***********************/

typedef struct capabilies_s {

  char *mechs;
  
  /* 0 = false; 1 = true */
  int starttls;
  int logindisabled;

} capabilities_t;

/*
 *
 */
char *
do_getuserid(globals_t *g)
{
    if (!g) {
	g = do_getcurrentg();
    }

    return g->userid;
}

/*
 * during connection need read
 */
int clie_cons_enable(pfconn_pt me,select_info_pt si)
{
  FD_SET(me->fd,&si->read_fd);
  return 0;
}

void clie_cons_data_sel(pfconn_pt me)
{
  char buf[500];
  register int rc;
  while(TRUE) {
    rc=read(me->fd,buf,sizeof(buf)-1);	/*read what we can*/
    if (rc == 0) rl_error("connection died");
    if(rc == -1) {
      if(errno==EWOULDBLOCK)break; /*nothing there*/
      er_ucall(rc,("read"));	/*make sure its not an error code*/
    }
    buf[rc] = '\0';
    CON_Write(buf);
  }
}

/*
 * accept a server data connection
 */
int clie_conn_sel(pfconn_pt me)
{
    printf("clie conn sel\n");
    return 0;
}

pfio_actions clie_cons_connect={
  serv_conn_enable,
  clie_conn_sel,
  never_sel,
  never_sel,
};

void  mn_connect_pt(c,port,host)
pfconn_pt c;
int port;
long host;
{
  struct sockaddr_in serv_sin;
  /* xxx   c->fd=new_socket(0); */
  bzero(&serv_sin,sizeof(serv_sin));
  serv_sin.sin_family=AF_INET;
  serv_sin.sin_port=htons(port);
  /*  if(host==0)
    host=get_netaddr(c->fd); */

  if((serv_sin.sin_addr.s_addr=host)==0)
    rl_error("can't get local host ip address ");
  if (port == GUARDIAN_PORT)
      printf("connecting to port %d",ntohs(serv_sin.sin_port));
  er_ucall(connect(c->fd,(struct sockaddr *)&serv_sin,sizeof(serv_sin)),("connect"));
}

/* 
 * client starts up connection 
 */

void startup_incomming(PFM_pt pkt)
{
    printf("startup incoming\n");
}

int sasl_cb_getsecret(sasl_conn_t *conn, void *context, int id,
		      sasl_secret_t **psecret)
{
    const char *server = (const char *)context;
    char *pass = NULL;

    int r = password_get(server,&pass);

    if ((r) || (!pass)) {
	return SASL_FAIL;
    }

    *psecret = (sasl_secret_t *) malloc(sizeof(sasl_secret_t)+strlen(pass)+1);
    if (!*psecret) return SASL_NOMEM;

    (*psecret)->len = strlen(pass);
    strcpy( (*psecret)->data, pass);

    return SASL_OK;
}

int sasl_cb_getuser(void *context, int id,
		    const char **result, unsigned *len)
{
    char *user;

    if(context) {
	user = (char *)context;
    } else { 
	return SASL_FAIL;
    }

    *result = xstrdup(user);
    if (!*result) return SASL_NOMEM;
    if (len) *len = strlen(user);

    return SASL_OK;
}

/*
 * IMAP command completion callback
 */
static void
callback_finish(struct imclient *imclient,
                void *rock,
                struct imclient_reply *reply)
{
    globals_t *g = (globals_t *)rock;
    g->cmds_finished++;

    IMAP_RESPONSE ( reply );

    if (!strcmp(reply->keyword, "OK")) {

        g->cmd_done = IMAP_OK;
    } else if (!strcmp(reply->keyword, "NO")) {
	CON_Write("IMAP command failed with: NO ");
	CON_WriteLine(reply->text);
        g->cmd_done = IMAP_NO;
    }
    else if (!strcmp(reply->keyword, "BAD")) {
	CON_Write("IMAP command failed with: BAD ");
	CON_WriteLine(reply->text);
        g->cmd_done = IMAP_BAD;

    }
    else if (!strcmp(reply->keyword, "EOF")) {
        g->cmd_done = IMAP_EOF;
    }
    else {
        printf("Huh? [%s]\n",reply->keyword);
        g->cmd_done = IMAP_BAD;
    }
}


void
callback_list(struct imclient *imclient,
              void *rock,
              struct imclient_reply *reply)
{
    char *s, *end;
    char *mailbox, *attributes, *separator;
    int c;
    globals_t *g = (globals_t *)rock;

    IMAP_RESPONSE ( reply );
    s = reply->text;
    
    if (*s++ != '(') return;
    end = strchr(s, ')');
    if (!end) return;
    attributes = s;
    s = end;
    *s++ = '\0';

    if (*s++ != ' ') return;
    if (*s == 'N') {
        if (s[1] != 'I' || s[2] != 'L') return;
        separator = "";
        s += 3;
    }
    else if (*s == '\"') {
        s++;
        if (*s == '\\') s++;
        separator = s++;
        if (*s != '\"') return;
        *s++ = '\0';
    }

    if (*s++ != ' ') return;
    c = imparse_astring(&s, &mailbox);
    if (c != '\0') return;

    RL_set(FOLDER_RL, spot, mailbox,0, g);
    LST_InvalidateItem(FOLDER_RL, spot);
    LST_SetListSize(FOLDER_RL, spot+1, 0,1);
    spot++;

    for (s = attributes; (end = strchr(s, ' '))!=NULL; s = end+1) {
        *s = '\0';

    }

}

void
callback_changed_list(struct imclient *imclient,
		      void *rock,
		      struct imclient_reply *reply)
{
    char *s, *end;
    char *mailbox, *attributes, *separator;
    int c;
    skiplist *slist = (skiplist *)rock;
    changed_item_t *newitem;

    IMAP_RESPONSE ( reply );
    s = reply->text;
    
    if (*s++ != '(') return;
    end = strchr(s, ')');
    if (!end) return;
    attributes = s;
    s = end;
    *s++ = '\0';

    if (*s++ != ' ') return;
    if (*s == 'N') {
        if (s[1] != 'I' || s[2] != 'L') return;
        separator = "";
        s += 3;
    }
    else if (*s == '\"') {
        s++;
        if (*s == '\\') s++;
        separator = s++;
        if (*s != '\"') return;
        *s++ = '\0';
    }

    if (*s++ != ' ') return;
    c = imparse_astring(&s, &mailbox);
    if (c != '\0') return;

    /* add entry to skiplist */
    newitem = xmalloc(sizeof(changed_item_t));
    newitem->name = xstrdup(mailbox);
    newitem->unseen = -1;

    sinsert(slist,newitem);

}

static void
callback_exists(struct imclient *imclient,
	       void *rock,
	       struct imclient_reply *reply)
{
    globals_t *g = (globals_t *)rock;

    IMAP_RESPONSE ( reply );
    g->exists = reply->msgno;

}

static void
callback_select(struct imclient *imclient,
	       void *rock,
	       struct imclient_reply *reply)
{
    globals_t *g = (globals_t *)rock;
    char *s = reply->text;

    IMAP_RESPONSE ( reply );
    if (s[0]!='[') return;
    s++;

    if (strncasecmp(s,"UNSEEN ",7)==0) {
	s+=7;

	g->unseen = 0;
	while (isdigit(*s)) {
	    g->unseen = (g->unseen*10) + (*s)-'0';
	    s++;
	}
    }
       
}

/* Handle a bodystrcture call */
static void 
callback_bodystructure(struct imclient *imclient,
		       void *rock,
		       struct imclient_reply *reply)
{  
  IMAP_RESPONSE ( reply );
}

static int
getint(char *in, char **out)
{
    int ret = 0;
    while (isdigit(*in)) {
	ret = (ret*10) + (*in) - '0';
	in++;
    }

    *out = in;

    return ret;
}

static void
callback_quota(struct imclient *imclient,
	       void *rock,
	       struct imclient_reply *reply)
{
    globals_t *g = (globals_t *)rock;
    char *s = reply->text;

    IMAP_RESPONSE ( reply );
    s = strstr(s,"STORAGE");
    if (!s) return;

    s+=8;

    g->quotaused = getint(s, &s);
    
    s++;

    g->quotalimit = getint(s, &s);
       
}

/* return malloc'd string containing the address */
static char *parseaddr(char *s)
{
    char *p;
    int len;

    p = s;

    if (*p++ != '<') return 0;

    /* at-domain-list */
    while (*p == '@') {
	p++;
	if (*p == '[') {
	    p++;
	    while (isdigit((int) *p) || *p == '.') p++;
	    if (*p++ != ']') return 0;
	}
	else {
	    while (isalnum((int) *p) || *p == '.' || *p == '-') p++;
	}
	if (*p == ',' && p[1] == '@') p++;
	else if (*p == ':' && p[1] != '@') p++;
	else return 0;
    }
    
    /* local-part */
    if (*p == '\"') {
	p++;
	while (*p && *p != '\"') {
	    if (*p == '\\') {
		if (!*++p) return 0;
	    }
	    p++;
	}
	if (!*p++) return 0;
    }
    else {
	/* disallow plus addressing without giving a name */
	if (*p == '+') return 0;
	
	while (*p && *p != '@' && *p != '>') {
	    if (*p == '\\') {
		if (!*++p) return 0;
	    }
	    else {
		if (*p <= ' ' || (*p & 128) ||
		    strchr("<>()[]\\,;:\"", *p)) return 0;
	    }
	    p++;
	}
    }

    /* @domain */
    if (*p == '@') {
	p++;
	if (*p == '[') {
	    p++;
	    while (isdigit((int) *p) || *p == '.') p++;
	    if (*p++ != ']') return 0;
	}
	else {
	    while (isalnum((int) *p) || *p == '.' || *p == '-') p++;
	}
    }
    
    if (*p++ != '>') return 0;
    if (*p && *p != ' ') return 0;
    len = p - s;

    s = xstrdup(s);
    s[len] = '\0';
    return s;
}

static void parsefrom(char *in, char *out, int outlen)
{
    char *s = NULL;
    char *left;

    memset(out,'\0',outlen);

    if (strncasecmp(in, "from:", 5) == 0) {
	in+=5;

	/* skip by whitespace */
	while (isspace(*in)) {
	    in++;
	}
	
	/* if quoted */
	if (*in == '\"') {
	    char *end;
	    
	    in++;
	    end = strchr(in,'\"');
	    
	    if (end) {
		*end = '\0';
		strncpy(out,in,outlen);
		return;
	    }
	}

	left = strchr(in,'<');
	if ((left) && (left!=in)) {
	    int tmp;

	    left--;
	    while (isspace(*left)) {
		left--;
	    }

	    if (left-in < outlen) {
		tmp = left-in+1;
	    } else {
		tmp = outlen;
	    }
	    strncpy(out,in,tmp);
	    return;
	} else if (left) {
	    s = parseaddr(left);
	} else {
	    strncpy(out,in,outlen);
	    return;
	}
    }
	
    if (!s) {
	strncpy(out,"???",outlen);
	return;
    }
    
    strncpy(out,s,outlen);
    free(s);
}

static void parsesize(char *in, char *out, int outlen)
{
    long num;
    char *endptr;

    num = strtoul(in,&endptr,10);

    if ((num == ULONG_MAX)) {
	strncpy(out,"?",outlen);
	return;
    }

    if (num > 1024*1024*1000) {
	snprintf(out,outlen,"%-3.0fG",(1.0*num)/(1024*1024*1024*1.0));
	return;
    }
    if (num > 1024*1000*10) {
	snprintf(out,outlen,"%-3.0fM",(1.0*num)/(1024*1024*1.0));
	return;
    }
    if (num > 1024*1000) {
	snprintf(out,outlen,"%-2.1fM",(1.0*num)/(1024*1024*1.0));
	return;
    }
    if (num >= 1024*10) {
	snprintf(out,outlen,"%-3.0fK",(1.0*num)/(1024*1.0));
	return;
    }
    if (num > 1000) {
	snprintf(out,outlen,"%-2.1fK",(1.0*num)/(1024*1.0));
	return;
    }
    
    strncpy(out,in,outlen);
}

int didsearch = 0;

void
callback_fetch(struct imclient *imclient,
	       void *rock,
	       struct imclient_reply *reply)
{
    char *s, *end;

    char line[81];

    char flags[5];
    char thedate[101];
    char size[8];
    char from[25];
    char subject[31];

    int day;
    char month[5];
    int year;
    
    char data[2048] = { 0 };
    int rlspot;

    IMAP_RESPONSE ( reply );
    /* zero things */
    subject[0] = '\0';
    from[0] = '\0';

    if (didsearch!=-1) {
	rlspot = didsearch;
	didsearch++;
    } else {
	rlspot = reply->msgno-1;
    }

    strncpy(data, reply->text, sizeof(data)-1);

    s = data;

    if (*s++ != '(') {
	CON_WriteLine("Fetch response missing initial '('");
	return;
    }
    if (strncmp(s,"FLAGS (",7)!=0) {
	CON_WriteLine("Fetch response missing FLAGS (");
	return;
    }
    s+=7;

    /* get flags */
    end = strchr(s, ')');    
    if (!end) {
	CON_WriteLine("Fetch response: message flags messed up");
	return;
    }
    *end = '\0';
    
    strcpy(flags,"N  ");

    while((s!=NULL) && (s<end)) {
	if (strncasecmp(s,"\\seen",5)==0) {
	    flags[0]=' ';
	} else if (strncasecmp(s,"\\deleted",8)==0) {
	    flags[1]='D';
	} else if (strncasecmp(s,"\\answered",9)==0) {
	    flags[2]='A';
	} else if (strncasecmp(s,"\\draft",6)==0) {
	  flags[2] = 'W';
	}

	s++;
	s = strchr(s,'\\');
    }
    
    s = end+1;

    if (strncmp(s," INTERNALDATE ",14)!=0) {
	CON_WriteLine("Fetch response: INTERNALDATE not found");
	return;
    }
    s+=14;

    if (*s++ != '\"') {
	CON_WriteLine("Fetch response: expeced \"");
	return;
    }

    /* get date */
    end = strchr(s, '\"');    
    if (!end) {
	CON_WriteLine("Fetch response: can't find trailing \"");
	return;
    }
    *end = '\0';

    day=strtol(s,NULL,10);
    strncpy(month,s+3,3);
    month[3] = '\0';
    year=strtol(s+7,NULL,10);

    snprintf(thedate,sizeof(thedate)-1,"%2d-%s-%02d",day,month,year%100);

    s = end+1;

    if (strncmp(s," RFC822.SIZE ",13)!=0) {
	CON_WriteLine("Fetch response: expected RFC822.SIZE");
	return;
    }
    s+=13;    

    /* get size */
    end = strchr(s, ' ');    
    if (!end) {
	CON_WriteLine("Fetch response: can't find end of size");
	return;
    }
    *end = '\0';

    memset(size,'\0',sizeof(size));
    parsesize(s,size,sizeof(size)-1);

    s = end+1;
   
    /*    if (strncmp(s,"BODY[HEADER.FIELDS (SUBJECT FROM DATE TO CC)] ",45)!=0) {
	CON_WriteLine("Fetch response: expected BODY[HEADER.FIELDS");
	CON_WriteLine(s);
	return;
    }
    s+=45;
    */
    /* Was cyrus-specific */
    if ( strncmp ( s, "BODY[HEADER.FIELDS", 18 ) != 0 ) { 
      CON_WriteLine ( "Fetch response: expected BODY[HEADER.FIELDS" );
      CON_WriteLine ( s );
      return;
    }
    s += 18;
    /* Search for the closing bracket */
    s = strchr ( s, ']' );
    if ( s == NULL ) { 
      CON_WriteLine ( "Fetch response: missing closing bracket" );
      return;
    }
    
    /* go past eol */
    s = strchr(s,'\n');
    if (!s) {
	CON_WriteLine("Fetch response: can't find end of line");
	return;
    }
    s++;    

    while (1) {

	if (s[0]=='\0') break;
	if (s[0]==')') break;

	/* put null at end */
	
	end = strchr(s,'\n');
	if (!end) {
	    CON_WriteLine("Fetch response: expect EOL");
	    break;
	}
	
	/* JCS - placing a NULL at the end of each line 
	 * causes long To: and Cc: lines to be cut short. 
	 * Makes it quite difficult to reply to emails with
	 * several recipients.
	 */
	//	*end = '\0';
	
	//	if (((char)*(end-1))=='\r') {
	//    *(end-1) = '\0';
	//	}

	if (strncasecmp(s,"Date:",5)==0) {
	  *end = '\0';
	    if ( *(end-1) == '\r' )
	      *(end-1) = '\0';
	  RL_set_header(RL_DATE,rlspot,s+6);

	} else if (strncasecmp(s,"From:",5)==0) {
	  *end = '\0';
	    if ( *(end-1) == '\r' )
	      *(end-1) = '\0';
	    RL_set_header(RL_FROM,rlspot,s+6);
	    parsefrom(s, from, sizeof(from)-1);
	    from[sizeof(from)-1]='\0';

	} else if (strncasecmp(s,"To:",3)==0) {

	    end = strchr ( s + 4, ':' );
	    if ( end ) { 
	      while ( *end != '\r' )
	        end--;
	      *end = '\0';
	    }

	    RL_set_header(RL_TO,rlspot,s+4);

	} else if (strncasecmp(s,"Subject:",8)==0) {
	    s+=9;
	    if ( *(end-1) == '\r' )
	      *(end-1) = '\0';
	    *end = '\0';
	    strncpy(subject,s, sizeof(subject)-1);
	    subject[sizeof(subject)-1]='\0';
	    RL_set_header(RL_SUBJECT,rlspot,s);

	} else if (strncasecmp(s,"Cc:",3)==0) {
	    end = strchr ( s + 4, ':' );
	    if ( end ) { 
	      while ( *end != '\r' )
	        end--;
	      *end = '\0';
	    } else { 
	      /* JCS - these two lines fix it from going off and reading 
	       * garbage after an empty Cc: line
	       */
	      end = s + 4;
	      while ( *end && *end != '\n' && *end != '\r' )
		end++;
	      *end = '\0';
	    }

	    RL_set_header(RL_CC,rlspot,s+4);

	} else if ((isspace(s[0])) || (s[0]=='\r') || (s[0]=='\n') || (s[0]=='\0')) {
	    /* xxx ignore now */
	} else {
	  char * c;
	  char foo[20];
	    CON_WriteLine("Fetch response: line not understood");
	    CON_WriteLine(s);
	    c = s;
	    while ( *c ) { 
	      sprintf ( foo, " %x", *c );
	      CON_Write ( foo );
	      c++;
	    }
	    CON_WriteLine ( "Subject was:" );
	    CON_WriteLine ( subject );
	    
	}
	if ( !end ) 
	  break;
	s = end+1;
    }

    snprintf(line,sizeof(line),"%-3s %-9s %-30s %-24s(%s)",
	     flags, thedate,
	     subject,from,size);

    RL_set(CAPTION_RL,rlspot, line,reply->msgno, NULL);
    LST_InvalidateItem(CAPTION_RL, rlspot);   
}

void
callback_fetchheaders(struct imclient *imclient,
		      void *rock,
		      struct imclient_reply *reply)
{
    char **headers = (char **)rock;
    char *s = reply->text;
    char *end;
    
    IMAP_RESPONSE ( reply );
    /* skip first line */
    s = strchr(s, '\n');
    if (!s) return;
    s++;

    /* remove trailing ")\n" */
    end = strrchr(s-2, '\n');
    if (end) end = strrchr(s, '\r');
    if (end) *end = '\0';

    *headers = xmalloc(strlen(s)+1);
    strcpy(*headers, s);
}

int cur_bodyline;
int inmiddle;
int g_anyleft;
char * g_msgbuffer = NULL;
char * g_leftoverbuffer = NULL;
int g_msg_decoded;

#define CONTENT_HEADER "Content-Transfer-Encoding:"

static int
append_data_to_screen(char *s)
{
    char line[79];

    while(*s != '\0') {
	char *next;
	char *end;

	end = strchr(s,'\n');
	if (end) {
	    *end = '\0';

	    if (((char)*(end-1))=='\r') {
		*(end-1) = '\0';
	    }
	    next = end+1;
	} else {
	    end = s + strlen(s);
	    next = end;
	}
	
	while (s<end) {

	    if (inmiddle) {
		cur_bodyline--;
		RL_read(BODY_RL,cur_bodyline,line);
		strncat(line, s, sizeof(line)-1-strlen(line));
		inmiddle = 0;
	    } else {
		strncpy(line, s, sizeof(line)-1);
	    }

	    line[ sizeof(line) -1 ] = '\0';

	    if (g_rot13_mode) rot13_line(line);

	    RL_set(BODY_RL,cur_bodyline, line,0,NULL);
	    LST_SetListSize(BODY_RL, cur_bodyline+1, 0,0);
	    LST_InvalidateItem(BODY_RL, cur_bodyline);

	    cur_bodyline++;

	    s+= (sizeof(line)-1);
	}

	s = next;
    }

    return 0;
}


void
callback_fetchbody(struct imclient *imclient,
		   void *rock,
		   struct imclient_reply *reply)
{
    char *s;
    char *end;
    int len;
    char encoding[100];
    char decoded[4096];
    int decodedlen;
    int r;
    
    IMAP_RESPONSE ( reply );
    s = reply->text;

    /* JCS - the returns have been replaced with 'goto giveup' because
     * the program would get stuck in a loop of reading nothing that it
     * couldn't parse properly from the server.  
     *  
     * In any case, this code doesn't parse what Cyrus IMAP returns
     * sometimes, so this code ought to be fixed.  Might as well be 
     * liberal in what we "accept" from the server, though.
     */
    
    /* should get: (BODY[HEADER.FIELDS (Content-Transfer-Encoding)] {%num%} */
    s = strchr(s,'\n');
    if (!s) {
      /* printf("error 1! got [%s]\n",reply->text); */
	goto giveup;
    }
    s++;
    
    end = strchr(s,'\n');
    if (!end) {
      //printf("error 2!\n");
	goto giveup;
    }
    *end = '\0';
    if (*(end-1) == '\r') *(end-1) = '\0';

    encoding[0] = '\0';

    if (strncasecmp(s,CONTENT_HEADER,strlen(CONTENT_HEADER))==0) {
      int 
        i;
	s+=strlen(CONTENT_HEADER);
	strncpy(encoding,s,sizeof(encoding)-1);
        for ( i = 0; i < sizeof(encoding); i++ )
          encoding[i] = toupper ( encoding[i] );
    }

    s = strstr(end+1,"BODY");
    if (!s) {
	printf("error 3\n");
	goto giveup;
    }
    
    /* find length we've been given */
    end = strchr(s,'\n');
    if (!end) goto giveup;
    if (end[-2]!='}') goto giveup;
    
    s=end-3;
    while (*s != '{') {
	s--;
    }
    s++;
    len = 0;
    while (isdigit(*s)) {
	len = (len*10) + (*s)-'0';
	s++;
    }
    s = end+1;
    s[len] = '\0';

    if (strstr(encoding,"base64")) {	
	char decodein[TRANSFER_BLOCK_SIZE+5] = { 0 };
	int pos = 0, i, overshoot;

        /* JCS - prevent leftovers from being lost in the base64 decode */
        if ( g_leftoverbuffer ) { 
          strcpy ( decodein, g_leftoverbuffer );
          pos += strlen ( g_leftoverbuffer );
          free ( g_leftoverbuffer );
          g_leftoverbuffer = NULL;
        }

	for (i = 0; i < len; i++ ) {
	    if ((s[i]!='\r') && (s[i]!='\n')) {
		decodein[pos++] = s[i];
	    }
	}
        
        overshoot = ( pos - 1 ) % MIME_BASE64_BLOCK;

        if ( overshoot != 0 ) { 
          g_leftoverbuffer = calloc ( MIME_BASE64_BLOCK + 1, sizeof(char) );
          for ( i = 0; i < overshoot; i++ )
            g_leftoverbuffer[i] = decodein[pos - overshoot + i];
        }

	
        pos -= overshoot;
        decodein[pos]='\0';

	r = sasl_decode64(decodein, pos,
			  decoded,&decodedlen);
	if (r != SASL_OK) {
	    printf("base64 decoding error\n");
	    goto giveup;
	}
	
	decoded[decodedlen] = '\0';
	s = decoded;
    }else if ( strstr ( encoding, "quoted-printable" ) ) { 
    rfc2045_from_mime_quoted_printable ( s, decoded, TRANSFER_BLOCK_SIZE-1, TRUE );
    s = decoded;
  }

    append_data_to_screen(s);

    if (len==2048) {
      RL_set(BODY_RL,cur_bodyline+1,(char *) -1,0, NULL);
      LST_SetListSize(BODY_RL, cur_bodyline+2, 0,0);
      inmiddle = 1;
    } else {
	g_anyleft = 0;
    }
    
    return;

 giveup:
    g_anyleft = 0;
    return;
    
}

void
callback_fetchbody_to_buffer ( struct imclient * imclient,
			       void * rock,
			       struct imclient_reply * reply )
{
  char * s;
  char *end;
  int len;
  char encoding[100];
  char decoded[TRANSFER_BLOCK_SIZE] = { 0 };
  int decodedlen;
  int r;
  
  IMAP_RESPONSE ( reply );
  s = reply->text;

  s = strstr ( s, CONTENT_HEADER );
  if ( !s ) { 
    goto giveup;
  }
  end  = strchr ( s, '\n' );

  encoding[0] = '\0';

  if (strncasecmp(s,CONTENT_HEADER,strlen(CONTENT_HEADER))==0) {
    int
      i;
    s+=strlen(CONTENT_HEADER);
    strncpy(encoding,s,sizeof(encoding)-1);
    for ( i = 0; i < sizeof(encoding); i++ )
      encoding[i] = toupper ( encoding[i] );
  }

  s = strstr(end+1,"BODY");
  if (!s) {
    printf("error 3\n");
    goto giveup;
  }
    
  /* find length we've been given */
  end = strchr(s,'\n');
  if (!end) goto giveup;
  if (end[-2]!='}') goto giveup;

  s = strchr ( s, '{' );

  if ( s == NULL ) {
    printf ( "error 4\n" );
    goto giveup;
  }
  s++;
  
  len = 0;
  while (isdigit(*s)) {
    len = (len*10) + (*s)-'0';
    s++;
  }
  s = end+1;

  if ( s[len] != ')' ) { 
    printf ( "error 5\n" );
    goto giveup;
  }
  s[len] = '\0';

  if ( g_msgbuffer != NULL ) 
    CON_WriteLine ( "About to leak memory in callback_fetchbody_to_buffer" ); 

  g_msgbuffer = xmalloc ( TRANSFER_BLOCK_SIZE + 1 );
  if ( g_msgbuffer == NULL ) { 
    CON_WriteLine ( "Out of memory" );
    goto giveup;
  }

  if (strstr(encoding,"BASE64")) {	
    char decodein[TRANSFER_BLOCK_SIZE+5] = { 0 };
    int pos = 0, i, overshoot;
    
    /* JCS - prevent leftovers from being lost in the base64 decode */
    if ( g_leftoverbuffer ) { 
      strcpy ( decodein, g_leftoverbuffer );
      pos += strlen ( g_leftoverbuffer );
      free ( g_leftoverbuffer );
      g_leftoverbuffer = NULL;
    }
    
    //len = strlen ( s );

    for (i = 0; i < len; i++ ) {
      if ((s[i]!='\r') && (s[i]!='\n')) {
	decodein[pos++] = s[i];
      }
    }

    overshoot = ( pos ) % MIME_BASE64_BLOCK;
    
    if ( overshoot != 0 ) { 
      g_leftoverbuffer = calloc ( MIME_BASE64_BLOCK, sizeof(char) );
      for ( i = 0; i < overshoot; i++ ) { 
        g_leftoverbuffer[i] = decodein[pos - overshoot + i];
      }
    }
    
    pos -= overshoot;
    decodein[pos]='\0';

    r = sasl_decode64(decodein, pos,
		      decoded,&decodedlen);
    if (r != SASL_OK) {
      CON_WriteLine ("base64 decoding error");
      goto giveup;
    }

    decoded[decodedlen] = '\0';
    s = decoded;
  } else if ( strstr ( encoding, "QUOTED-PRINTABLE" ) ) { 
    rfc2045_from_mime_quoted_printable ( s, decoded, TRANSFER_BLOCK_SIZE-1, TRUE );
    s = decoded;
    decodedlen = strlen ( s );
  } else { 
    strcpy ( g_msgbuffer, s );
    decodedlen = strlen ( s );
  }

  memcpy ( g_msgbuffer, s, decodedlen );

  if ( decodedlen == (TRANSFER_BLOCK_SIZE-1) ) { 
    inmiddle = 1;
  } else { 
    g_anyleft = 0;
    inmiddle = 0;
  }


  g_msg_decoded += decodedlen;
  msg_read += len;

  return;

 giveup:
  g_anyleft = 0;
  inmiddle = 0;
  return;
}

typedef struct tuple_s {
    skiplist *slist;
    globals_t *g;
} tuple_t;

void
callback_status(struct imclient *imclient,
		void *rock,
		struct imclient_reply *reply)
{
    char *s = reply->text;
    
    tuple_t *tup = (tuple_t *)rock;
    skiplist *slist = tup->slist;
    globals_t *g = tup->g;
    int unseen;
    char *mailbox;
    changed_item_t *item;
    skipnode *snode;
    int c;

    IMAP_RESPONSE ( reply );

    if (!slist) return;

    /* get the mailbox IMAP is talking about */
    c = imparse_astring(&s, &mailbox);
    if (c != ' ') return;

    /* find it in our list */
    for (item = (changed_item_t *)sfirst(slist,&snode); item != NULL; 
	 item = (changed_item_t *)snext(&snode) ) {
	if (strcmp(item->name,mailbox)==0) break;
    }

    if (!item) return;

    /* find the unseen number IMAP said */
    s = strchr(s,'(');
    if (!s) return;
    s++;
    if (strncasecmp(s,"UNSEEN ",7)!=0) return;
    s+=7;
    
    unseen = 0;
    while (isdigit(*s)) {
	unseen = (unseen*10) + (*s)-'0';
	s++;
    }

    
    item->unseen = unseen;

    /* If is INBOX display this info in the console */
    if (strcasecmp(item->name,"inbox")==0) {
	char tmpstr[100];
	
	snprintf(tmpstr,sizeof(tmpstr)-1,"Inbox has %d unread messages",item->unseen);
	CON_WriteLine(tmpstr);
    }

    /* if has unseen messages or is inbox add the window */
    if ((item->unseen > 0) || (strcasecmp(item->name,"inbox")==0)) {
	RL_set(FOLDER_RL, spot, item->name,0,g);
	LST_InvalidateItem(FOLDER_RL, spot);
	LST_SetListSize(FOLDER_RL, spot+1, 0,1);
	spot++;	    
    }
}

void
callback_status_unseen(struct imclient *imclient,
		       void *rock,
		       struct imclient_reply *reply)
{
    char *s = reply->text;
    
    int *unseen = (int *)rock;
    char *mailbox;
    int c;

    IMAP_RESPONSE ( reply );

    c = imparse_astring(&s, &mailbox);
    if (c != ' ') return;

    s = strchr(s,'(');
    if (!s) return;
    s++;
    if (strncasecmp(s,"UNSEEN ",7)!=0) return;
    s+=7;
    
    *unseen = 0;
    while (isdigit(*s)) {
	*unseen = ((*unseen)*10) + (*s)-'0';
	s++;
    }
}

static capabilities_t *parsecapabilitylist(char *str)
{
  char *tmp;
  int num=0;
  capabilities_t *ret=(capabilities_t *) xmalloc(sizeof(capabilities_t));
  ret->mechs = (char *)xmalloc(strlen(str)+1);
  ret->starttls=0;
  ret->logindisabled=0;

  /* check for stattls */
  if (strstr(str,"STARTTLS")!=NULL)
  {
    ret->starttls=1;
  }

  /* check for login being disabled */
  if (strstr(str,"LOGINDISABLED")!=NULL)
  {
    ret->logindisabled=1;
  }

  strcpy(ret->mechs,"");

  while ((tmp=strstr(str,"AUTH="))!=NULL)
  {
    char *end=tmp+5;
    tmp+=5;

    while(((*end)!=' ') && ((*end)!='\0'))
      end++;

    (*end)='\0';

    /* add entry to list */
    if (num>0)
      strcat(ret->mechs," ");
    strcat(ret->mechs, tmp);
    num++;

    /* reset the string */
    str=end+1;

  }

  return ret;
}

/*
 * Callback for OK. Look for [ALERT]'s
 */ 

static void callback_alert(struct imclient *imclient, 
			   void *rock,
			   struct imclient_reply *reply)

{
    char *s;
    IMAP_RESPONSE ( reply );
    s = reply->text;
 
    if (strncasecmp(s,"[ALERT]",7)==0) {
	DLG_Alert(ALERT_warning,s);
    }
}

/*
 * IMAP command completion callback
 */
static void callback_capability(struct imclient *imclient, 
				void *rock,
				struct imclient_reply *reply)

{
    char *s;
    capabilities_t **caps = (capabilities_t **) rock;
    
    IMAP_RESPONSE ( reply );
    s = reply->text;
 
    *caps = parsecapabilitylist(s);
}

static int 
do_disconnect(globals_t *g)
{
    if (g->imclient_conn == NULL) return 0;

    if ((g->selected_mbox) && (strncasecmp(g->selected_mbox,"inbox",5) == 0) && (config_getswitch("AutoExpunge",0))) {
	do_expunge(g);
    }
    
    imclient_close(g->imclient_conn);
    g->imclient_conn = NULL;

    return 0;
}

int
disconnect_all(void)
{
    int i;

    for (i = 0; i < num_globs; i++) {
	int r = do_disconnect(globs[i]);
	if (r) return r;
	globs[i] = NULL;
    }

    num_globs = 0;

    LST_SetListSize(FOLDER_RL, 0L, 0L,0);
    LST_Show(FOLDER_RL, TRUE);

    return 0;
}

static int mn_clie_connect(globals_t *g)
{
    int r = 0;
    sasl_callback_t cbs[5];
    capabilities_t *capabilitylist;
    int usetls = config_getswitch("imaps", 1);

    cbs[0].id = SASL_CB_PASS;
    cbs[0].proc = &sasl_cb_getsecret;
    cbs[0].context = g->server_hostname;
    cbs[1].id = SASL_CB_AUTHNAME;
    cbs[1].proc = &sasl_cb_getuser;
    cbs[1].context = g->userid;
    cbs[2].id = SASL_CB_USER;
    cbs[2].proc = &sasl_cb_getuser;
    cbs[2].context = g->userid;
    cbs[3].id = SASL_CB_LIST_END;

    set_net_is_connecting(TRUE);

    /* connect to the server */
    if (usetls) {
	r = imclient_connect (&g->imclient_conn, g->server_hostname, "imaps", (sasl_callback_t *)&cbs);
	if (r != 0) {
	    r = imclient_connect (&g->imclient_conn, g->server_hostname, "imap", (sasl_callback_t *)&cbs);
	}
    } else {
	r = imclient_connect (&g->imclient_conn, g->server_hostname, "imap", (sasl_callback_t *)&cbs);
    }

    if (r) {
	DLG_Alert(ALERT_warning,"Unable to connect to remote host");
	return -1;
    }

    /* add alert callback */
    imclient_addcallback(g->imclient_conn,
			 "OK", 0, callback_alert,
			 (void *)0, (char *)0);

    /* get the capabilities */
    imclient_addcallback(g->imclient_conn, "CAPABILITY", 0,
			 callback_capability, (void *) &capabilitylist, 
			 (char *) 0);

    g->cmd_done = NOTFINISHED;
    imclient_send(g->imclient_conn, callback_finish, (void *) g,
		  "CAPABILITY");

    while (g->cmd_done == NOTFINISHED) {
	imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done != IMAP_OK) {
	CON_WriteLine("Capability failed");
	return -1;
    }

#ifdef HAVE_SSL
    if (capabilitylist->starttls) {
	int layer;
	char layerstr[100];

	r = imclient_starttls(g->imclient_conn,10,"","",&layer);
	if (r) {
	    CON_WriteLine("Error starting TLS. Going to keep going anyway");
	}
	snprintf(layerstr,sizeof(layerstr)-1,"Negotiated TLS layer of strength %d",layer);
	CON_WriteLine(layerstr);
    }
#endif /* HAVE_SSL */

    r = imclient_authenticate(g->imclient_conn,
			      capabilitylist->mechs,
			      "imap",
			      g->userid,
			      0, 0);
    if (r) {
	char *password;
        
    loginretry:
        
	r = password_get(g->server_hostname,&password);

	if ((r) || (!password)) {
	    CON_WriteLine("LOGIN failed: Can't login without password");
	    return -1;
	}

	/* try the LOGIN method */
	g->cmd_done = NOTFINISHED;
	imclient_send(g->imclient_conn, callback_finish, (void *)g,
		      "LOGIN %s %s", 
		      g->userid,password);

	
	while (g->cmd_done == NOTFINISHED) {
	    imclient_processoneevent(g->imclient_conn);
	}
	
	if (g->cmd_done!=IMAP_OK) {
          char buf[STR255LEN];
          
          snprintf ( buf, STR255LEN, "Login to %s as %s failed.  Try again?",
                     g->server_hostname, 
                     g->userid );
          
          if ( DLG_GetBooleanFromUser ( buf,
                                        TRUE ) ) { 
            password_purge ( g->server_hostname );
            goto loginretry;
          } else { 
	    return -1;
          }
	}
		
    }

    set_net_dead(FALSE);
    appl_ConnectHappened();

    if (capabilitylist) {
	free(capabilitylist->mechs);
	free(capabilitylist);
    }

    return 0;
}

int mn_clie_init(char *host, char *user, globals_t **glob)
{
    int r;
    globals_t *g = xmalloc(sizeof(globals_t));
    memset(g,0,sizeof(globals_t));

    g->server_hostname = xstrdup(host);
    g->userid = xstrdup(user); 
    g->inboxunseen = -1;

    r = mn_clie_connect(g);
    if (r) { 
      free ( g->userid );
      free ( g->server_hostname );
      free ( g );
      return r;
    }

    globs[num_globs++] = g; /* xxx what if fails */
   
    *glob = g;

    currentmbox_g = g;

    return 0;
}

char *do_get_selected_mbox(globals_t *g)
{
    return g->selected_mbox;
}

/* 
 * So you can figure out whether a message in this bboard has been
 * deleted in this session or not 
 */
int do_get_deletedMessages ( globals_t * g )
{
    return g->deletedMessages;
}


static int do_select(globals_t *g, char *mboxname)
{
    int trys = 0;

    if ((g->selected_mbox) && (strncasecmp(g->selected_mbox,"inbox",5) == 0) && (config_getswitch("AutoExpunge",0))) {
	do_expunge(g);
    }

    RESTART_SPOT

    /* select mailbox */
    imclient_addcallback(g->imclient_conn,
			 "EXISTS", CALLBACK_NUMBERED, callback_exists,
			 (void *)g, (char *)0);
    imclient_addcallback(g->imclient_conn,
			 "OK", 0, callback_select,
			 (void *)g, (char *)0);

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "%s %s", "SELECT", mboxname);



    while (g->cmd_done == NOTFINISHED) {
	imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done == IMAP_OK) {
	if (g->selected_mbox) free(g->selected_mbox);
	g->selected_mbox = xstrdup(mboxname);
        /* New mailbox, so we don't have any deleted messages now */
	g->deletedMessages = FALSE;
    }

    currentmbox_g = g;

    return g->cmd_done;
}

int puntmailbox(globals_t *g, char *mbox)
{
    char *cur_mbox;
    int r;
    int iscurmbox = 0;
    
    if (!mbox) return 0;
    
    cur_mbox = do_get_selected_mbox(g);

    cur_mbox = xstrdup(cur_mbox);
    
    if (strcmp(cur_mbox,mbox)==0) iscurmbox = 1;
    
    if (!iscurmbox) {
	r = do_select(g, mbox);
	if (r!=IMAP_OK) return -1;
    }

    r = do_mark_all_seen(g);	  

    if (iscurmbox) {
	int i;
	int num = GET_LIST_SIZE(CAPTION_RL);
	for (i=0;i<num;i++) {
	    RL_setchar(CAPTION_RL,i,0,' ');
	}
	WIN_RedrawScreen();	  
    }
	
    if (!iscurmbox) {
	do_select(g, cur_mbox);
    }

    free(cur_mbox);

    return r;
}

int do_set_answered(globals_t *g, char *mbox, int spot)
{
    char *cur_mbox;
    int r;

    if (!mbox) return 0;
    
    cur_mbox = do_get_selected_mbox(g);

    if ((!cur_mbox) || (strcmp(mbox,cur_mbox))) {
	cur_mbox = xstrdup(cur_mbox);

	r = do_select(g, mbox);
	if (r!=IMAP_OK) return -1;

	r = do_setflag(g, rl_getmsgno(CAPTION_RL,spot),FLAG_ANSWERED);
	
	do_select(g, cur_mbox);
	free(cur_mbox);

	return r;
    } else {
	
	/* only put A if current mbox */
	RL_setchar(CAPTION_RL,spot,2,'A');
	LST_InvalidateItem(CAPTION_RL, spot);

	return do_setflag(g, rl_getmsgno(CAPTION_RL,spot),FLAG_ANSWERED);
    }
}

static int ch_compare(const void *a, const void *b)
{
    changed_item_t *aa = (changed_item_t *)a;
    changed_item_t *bb = (changed_item_t *)b;

    return strcmp(aa->name,bb->name);
}

int do_showchanged(globals_t *g)
{
    skiplist *slist;
    skipnode *snode;
    changed_item_t *item;
    int sent = 0;
    int trys = 0;

    RESTART_SPOT

    slist = newlist(10, 0.5, (int (*)(const void *, const void *)) &ch_compare);
    
    imclient_addcallback(g->imclient_conn,
                         "LSUB",
			 0, callback_changed_list,
                         (void *)slist, (char *)0);

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
                  "LSUB \"*\" \"*\"" );

    g->cmd_done = NOTFINISHED;

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) {
	goto restart;
    }

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("LSUB command failed\n");
	return -1;
    }

    if (sempty(slist)) {
	DLG_Alert(ALERT_warning,"You are not subscribed to any bboards. You probably want to subscribe to your inbox. Select Expose from BBoards/Folders menu to see bboards");
	return 0;
    }

    if (!sempty(slist)) {
	tuple_t tup;
	tup.slist = slist;
	tup.g = g;

	imclient_addcallback(g->imclient_conn,
			     "STATUS", 0, callback_status,
			     (void *)&tup, (char *)0);
	
	g->cmds_finished = 0;
	
	g->cmd_done = NOTFINISHED;
	for (item = (changed_item_t *)sfirst(slist,&snode); item != NULL; 
	     item = (changed_item_t *)snext(&snode) ) {
	    
	    imclient_send(g->imclient_conn, callback_finish, (void *)g,
			  "STATUS %s (UNSEEN)", item->name);
	    sent++;
	}
	
	
	while ((g->cmd_done == NOTFINISHED) || (g->cmds_finished < sent)) {
	    imclient_processoneevent(g->imclient_conn);

	    if (g->cmd_done == IMAP_EOF) {
		goto restart;
	    }
	} 	
    }

    /* free all items */
    for (item = (changed_item_t *)sfirst(slist,&snode); item != NULL; 
	     item = (changed_item_t *)snext(&snode) ) {
	    
	free(item->name);
	free(item);
    }
    destroylist(slist);


    return 0;
}

int do_all_showchanged(void)
{
    int i;
    int r = 0;

    /* clear list */
    spot = 0;
    LST_SetListSize(FOLDER_RL, 0L, 0L,0);

    for (i = 0; i < num_globs; i++) {
	r = do_showchanged(globs[i]);
    }

    LST_Show(FOLDER_RL, TRUE);

    return r;
}

int do_showlist(globals_t *g,
		int type, /* 0 - list 1 - lsub */
		char *pattern)
{
    char *list_type;
    int trys = 0;

    RESTART_SPOT

    if (type == 0) {
	list_type = "LIST";
    } else {
	list_type = "LSUB";
    }

    imclient_addcallback(g->imclient_conn,
                         list_type,
			 0, callback_list,
                         (void *)g, (char *)0);

    g->cmd_done = NOTFINISHED;
    imclient_send(g->imclient_conn, callback_finish, (void *)g,
                  "%s %s %s", 
		  list_type,
		  pattern,
		  "*" );


    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("List failed");
	return -1;
    }

    return 0;
}

int do_all_showlist(int type, /* 0 - list 1 - lsub */
		    char *pattern)
{
    int i;
    int r = 0;

    spot = 0;
    LST_SetListSize(FOLDER_RL, 0L, 0L,0);

    for (i = 0; i < num_globs; i++) {
	r = do_showlist(globs[i], type, pattern);
    }

    LST_Show(FOLDER_RL, TRUE);

    return r;
}

static void
callback_search(struct imclient *imclient,
		void *rock,
		struct imclient_reply *reply)
{
    char *s = reply->text;
    skiplist *slist = (skiplist *)rock;
    char numstr[10];
    
    IMAP_RESPONSE ( reply );

    while (*s != '\0') {
	int spot = 0;
	long num;
	set_t *newset;

	while(isdigit(*s)) {
	    numstr[spot++] = *s;
	    s++;
	}
	numstr[spot] = '\0';

	num = strtoul(numstr, NULL, 10);
	newset = xmalloc(sizeof(set_t));
	newset->low = num;
	newset->high = num;

	sinsert(slist, newset);

	if (*s == ' ') {
	    s++;
	}
    }
}
/*
 * Show the quota usage
 *
 */
static int
do_showquota(globals_t *g)
{
    char msg[1024];
    int trys  = 0;
    
    RESTART_SPOT
    
    g->quotaused = 0;
    g->quotalimit = -1;

    imclient_addcallback(g->imclient_conn,
                         "QUOTA", 0, callback_quota,
                         (void *)g, (char *)0);

    g->cmd_done = NOTFINISHED;
    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "GETQUOTAROOT INBOX");
    

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Quota fetch failed of bboard failed");
	return -1;
    }

    if (g->quotalimit <= 0) {
	snprintf(msg,sizeof(msg)-1,"Quota for %s: unlimited", g->server_hostname);
    } else {
	float p = (g->quotaused * 1.0) / (g->quotalimit * 1.0);
	int perc = (p * 100);

	snprintf(msg,sizeof(msg)-1,"Quota for %s: %d%% used (%d of %d)",
		 g->server_hostname, perc,
		 g->quotaused, g->quotalimit);
    }
    DLG_Alert(ALERT_warning,msg);
    
    return 0;
}

int do_all_showquota(void)
{
    int i;
    int r = 0;

    for (i = 0; i < num_globs; i++) {
	r = do_showquota(globs[i]);
    }

    return r;
}


#define AMOUNT 100

static char *
setlist_to_str(skiplist *slist, int *size)
{
    set_t *item;
    skipnode *snode;
    char *ret = xmalloc(AMOUNT);
    int alloclen = AMOUNT;
    int num = 0;
    strcpy(ret,"");


    for (item = (set_t *)sfirst(slist,&snode); item != NULL; 
	 item = (set_t *)snext(&snode) ) {

	num++;
	
	if (strlen(ret) + 10 < alloclen) {
	    alloclen+=AMOUNT;
	    ret = xrealloc(ret,alloclen);
	}

	if (strlen(ret) != 0) {
	    strcat(ret,",");
	}
	
	snprintf(ret+strlen(ret),10,"%d",item->low);	
    }    

    if (size) *size = num;

    return ret;
}

static int set_compare(const void *a, const void *b)
{
    set_t *aa = (set_t *)a;
    set_t *bb = (set_t *)b;

    if (aa->low < bb->low) return -1;
    return 1;
}

skiplist *do_search(globals_t *g, char *crit)
{
    skiplist *slist;
    int trys = 0;

    if (0) {    
    restart: 
	trys++; 
	if (trys > 2) return NULL;
	CON_WriteLine("Connection closed. Reconnecting..."); 
	mn_clie_connect(g); 
    } 


    slist = newlist(100, 0.5, (int (*)(const void *, const void *)) &set_compare);

    imclient_addcallback(g->imclient_conn,
                         "SEARCH", 0, callback_search,
                         (void *)slist, (char *)0);

    g->cmd_done = NOTFINISHED;
    imclient_send(g->imclient_conn, callback_finish, (void *)g,
                  "SEARCH %a",crit);
    

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) return NULL;

    return slist;
}

/* where the list of ones we've fetched starts */
int start_list;
int end_list;

int do_fetch_more(globals_t *g, int spot)
{
#define MSGS_TO_GET 25 
    int start;
    int end;
    int trys = 0;

    RESTART_SPOT

    /* this seems to get called from within do_fetch sometimes. that's bad! */
    if (g->cmd_done == NOTFINISHED) return -1;
    
    if ((spot > start_list) && (spot < end_list)) {
	return 0;
    }

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, callback_fetch,
                         (void *)0, (char *)0);

    if (spot <= start_list) {

	start = start_list - MSGS_TO_GET;

	if (spot < start) {
	    start = spot - (MSGS_TO_GET/2);
	}

	if (start < 1) {
	    start = 1;
	}
	end = start_list-1;

    } else {
	
	start = end_list;
	
	end = end_list + MSGS_TO_GET;

	if (spot > end) {
	    end = spot + (MSGS_TO_GET/2);
	}
	
	if (end > g->exists) {
	    end = g->exists;
	}
    }

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
                  "FETCH %d:%d (FLAGS INTERNALDATE RFC822.SIZE BODY.PEEK[HEADER.FIELDS (SUBJECT FROM DATE TO CC)])",start, end);
    
    if (start < start_list) {
	start_list = start;
    }

    if (end > end_list) {
	end_list = end;
    }


    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Fetch failed (do_fetch_more)");
	return -1;
    }

    return 0;
}

int do_fetch(globals_t *g, char *mboxname, char *crit)
{
     int r;
     int start;
     int highlight;
     int end;
     skiplist *slist = NULL;
     char *range;
     int trys = 0;

     RESTART_SPOT

     g->unseen = -1;
     didsearch = -1;

    r = do_select(g, mboxname);
    if (r != IMAP_OK) {
	CON_WriteLine("Select of mailbox failed");
	return -1;
    }

    if (crit) {
	slist = do_search(g, crit);	
	if (slist == NULL) {
	    return 0;
	}
	didsearch = 0;    
    }

    LST_SetListSize(CAPTION_RL, 0L, 0L,0);
    RL_zero(CAPTION_RL);

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, callback_fetch,
                         (void *)0, (char *)0);

    if (slist == NULL) {
	if (g->unseen==-1) {
	  /* JCS - this added 10 to the start value.  Does that make sense?
	   * If nothing else, it made a nice segfault.
	   */
	  start = g->exists - GET_LIST_pt(CAPTION_RL)->win->awin->data_height;
	    highlight = g->exists-1;
	} else {
	    start = g->unseen-5;
	    highlight = g->unseen -1;
	}
	
	if (start < 1) start = 1;
	
	end = start+GET_LIST_pt(CAPTION_RL)->win->awin->data_height + 10;
	if (end>g->exists) end = g->exists;

	range = xmalloc(30+1);
	snprintf(range,30,"%d:%d",start,end);
    } else {
	start = 0;
	range = setlist_to_str(slist, &end);
	highlight = end -1;	

	if (end == 0) {
	    DLG_Alert(ALERT_warning,"Search produced no results");
	    return 0;
	}
    }

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
                  "FETCH %s (FLAGS INTERNALDATE RFC822.SIZE BODY.PEEK[HEADER.FIELDS (SUBJECT FROM DATE TO CC)])",range);
    
    free(range);

    if (slist == NULL) {
	LST_SetListSize(CAPTION_RL, g->exists, start,1);
    } else {
	LST_SetListSize(CAPTION_RL, end, start,1);
    }

    start_list = start;
    end_list = end;


    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Fetch failed (do_fetch)");
	return -1;
    }

    LST_Show(CAPTION_RL, TRUE);
    if (highlight == -1) highlight = 0;
    LST_SetSelection(CAPTION_RL, (long)highlight);

    return 0;
}

static void add_header(const char *name, int type, int msgno)
{
    char line[80];
    char *header;

    header = RL_get_header(type,msgno);
    if (!header) return;
    if (*header == '\0') return;

    snprintf(line,sizeof(line),"%s: %s",name,header);

    RL_set(BODY_RL,cur_bodyline,line,0, NULL);    
    LST_InvalidateItem(BODY_RL, cur_bodyline);

    cur_bodyline++;
}

static void
show_headers(globals_t *g, int spot)
{

    if (!g_fullheaders_mode) {
	add_header("Date",   RL_DATE   , spot);
	add_header("To",     RL_TO     , spot);
	add_header("Cc",     RL_CC     , spot);
	add_header("From",   RL_FROM   , spot);
	add_header("Subject",RL_SUBJECT, spot);
    } else {
	char *headers = do_getheaders(g, spot);

	append_data_to_screen(headers);
    }
}

int do_anyleft(void)
{
    return g_anyleft;
}

char* do_getheaders(globals_t *g, int spot)
{
    int msgno = rl_getmsgno(CAPTION_RL,spot);
    char *headers;
    int trys = 0;

    if (0) {    
    restart: 
	trys++; 
	if (trys > 2) return NULL;
	CON_WriteLine("Connection closed. Reconnecting..."); 
	mn_clie_connect(g); 
    } 

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, callback_fetchheaders,
                         (void *)&headers, (char *)0);

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "FETCH %d body[0]",msgno);


    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Error retrieving headers");
	return NULL;
    }

    return headers;
}

int do_getmsg(globals_t *g, int spot)
{
    int msgno = rl_getmsgno(CAPTION_RL,spot);
    int trys = 0;

    RESTART_SPOT

    /* turn off the 'N' */
    RL_setchar(CAPTION_RL,spot,0,' ');
    LST_InvalidateItem(CAPTION_RL, spot); /* tell it to redraw */

    cur_bodyline = 0;
    inmiddle = 0;
    g_anyleft = 1;

    /* insert the headers */
    show_headers(g, spot);

    /* add an empty line */
    RL_set(BODY_RL,cur_bodyline,"",0, NULL);    
    LST_InvalidateItem(BODY_RL, cur_bodyline);
    cur_bodyline++;

    LST_SetListSize(BODY_RL, cur_bodyline, 0,1);

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, callback_fetchbody,
                         (void *)0, (char *)0);

    g->cmd_done = NOTFINISHED;

    /* JCS - I want to see the entire message, so that I can tell whether
     * there is an attachment or not.
     */
    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "FETCH %d (BODY[HEADER.FIELDS (Content-Transfer-Encoding)] body[TEXT]<0.2048>)",msgno);

    msg_read = 2048;

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Fetch failed (do_getmsg)");
	return -1;
    }

    do_msg_banner ( g, spot );

    return 0;
}

/* JCS - set the title banner of the message window to have excerpts of the
 * subject and sender
 */
int do_msg_banner(globals_t *g, int spot )
{
  char 
    * subject,
    * from,
    banner[STR255LEN] = { 0 };

  int
    len;

  subject = RL_get_header ( RL_SUBJECT, spot );
  from    = RL_get_header ( RL_FROM, spot );

  if ( subject == NULL )
    subject = "";
  
  if ( from == NULL ) 
    from = "";

  banner[0] = ' ';
  strncpy ( banner + 1, subject, 35 );
  len = imin ( 35, strlen ( subject ) );

  banner[len+1] = ':';
  banner[len+2] = ' ';

  strncpy ( banner+len+3, from, 35 );
  banner[75] = 0;

  LST_SetBanner ( BODY_RL, banner );
  
  return 0;
}

int do_getmsg_more(globals_t *g, int spot)
{
    int msgno = rl_getmsgno(CAPTION_RL,spot);
    int trys = 0;

    RESTART_SPOT

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, callback_fetchbody,
                         (void *)0, (char *)0);

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "FETCH %d (BODY[HEADER.FIELDS (Content-Transfer-Encoding)] body[TEXT]<%d.2048>)",msgno,msg_read);

    msg_read+=2048;


    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Error fetching more of message");
	return -1;
    }

    return 0;
}

int do_getmsg_to_buffer(globals_t *g, int spot, const char * attachment, int startbyte)
{
    int msgno = rl_getmsgno(CAPTION_RL,spot);
    int trys = 0;

    if ( g_msgbuffer != NULL ) 
      free ( g_msgbuffer );

    g_msgbuffer = NULL;

    RESTART_SPOT

      msg_read = 0;
    g_msg_decoded = 0;

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, 
			 callback_fetchbody_to_buffer,
                         (void *)0, (char *)0);

    g->cmd_done = NOTFINISHED;

    if ( strcasecmp ( attachment, "TEXT" ) ) { 
      imclient_send(g->imclient_conn, callback_finish, (void *)g,
                    "FETCH %d (BODY[%s.MIME] body[%s]<%d.%d>)",
                    msgno, attachment, attachment, startbyte, TRANSFER_BLOCK_SIZE-1);
    } else { 
      imclient_send(g->imclient_conn, callback_finish, (void *)g,
                    "FETCH %d (BODY[HEADER.FIELDS (Content-Transfer-Encoding)] body[%s]<%d.%d>)",
                    msgno, attachment, startbyte, TRANSFER_BLOCK_SIZE-1);
    }
    

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Error fetching more of message");
	return -1;
    }

    return msg_read;
}

int do_save_to_file ( const int    fd,
		      const char * attachmentNumber )
{
  LIST_pt alist;
  LIST save_list;
  int  pos = 0, lastPos = 0;
  int  list_no = LST_GetListidFromWindow();
  int  len;
  char * attachment;

  if (list_no <= 0 || (alist = GET_LIST_pt(list_no)) == (LIST_pt)NULL)
    return(FALSE);

  if (alist->size <= 0) return(TRUE);

  SCR_ClearKeyboard();
  bcopy(alist, &save_list, sizeof(LIST));

  msg_read = 0;

  if ( attachmentNumber == NULL || 
       strlen ( attachmentNumber ) == 0 ) 
    attachment = "TEXT";
  else
    attachment = (char*)attachmentNumber; 

  do { 
    CON_ShowTXStatus ( msg_read / 1024 );

    lastPos = pos;

    pos =  do_getmsg_to_buffer ( do_getcurrentg(), 
				 GET_LIST_SELECTION ( CAPTION_RL ),
				 attachment,
				 pos );

    /* If we're downloading a large message, you might want the user
     * to cancel the operation in the middle.
     */
    if (SCR_Keyboard_io() && 
	SCR_ClearKeyboard() &&
	DLG_GetBooleanFromUser(MSG_ServerWait, FALSE)) {
      break;
    }
       

    if ( pos <= 0 || g_msgbuffer == NULL ) 
      break;

    len = strlen ( g_msgbuffer );
    if ( write ( fd, g_msgbuffer, pos - lastPos ) != len ) { 
      CON_WriteLine ( "Error writing attachment to output pipe!" );
      break;
    }
  } while ( 1 );
  
  /* Clear the transfer status */
  CON_ShowTXStatus ( 0 );
  
  bcopy(&save_list, alist, sizeof(LIST));
  WIN_RedrawScreen();

  return TRUE;
}

int do_getbodystructure(globals_t * g,
			int num )
{
  
    int msgno = rl_getmsgno(CAPTION_RL,num);
    int trys = 0;

    RESTART_SPOT

    g->cmd_done = NOTFINISHED;

    imclient_addcallback(g->imclient_conn,
                         "FETCH", CALLBACK_NUMBERED, 
			 callback_bodystructure,
                         (void *)0, (char *)0);

    g->cmd_done = NOTFINISHED;
    imclient_send ( g->imclient_conn, callback_bodystructure,
		    (void *)g, "FETCH %d (BODYSTRUCTURE)",
		    msgno );

    while (g->cmd_done == NOTFINISHED) {
      imclient_processoneevent(g->imclient_conn);
    }
    
    if (g->cmd_done == IMAP_EOF) goto restart;
    
    if (g->cmd_done!=IMAP_OK) {
      CON_WriteLine("Fetching bodystructure failed" );
      return -1;
    }

    return TRUE;
}


int do_subscribe(globals_t *g, char *mboxname, int sub) /* 0 - unsub 1 - sub */
{
    int trys = 0;

    RESTART_SPOT

    g->cmd_done = NOTFINISHED;

    if (sub == 1) {
	imclient_send(g->imclient_conn, callback_finish, (void *)g,
		      "SUBSCRIBE %s",mboxname);
    } else {
	imclient_send(g->imclient_conn, callback_finish, (void *)g,
		      "UNSUBSCRIBE %s",mboxname);
    }

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Subscribe change failed");
	return -1;
    }

    return 0;
}

int do_flag(globals_t *g, int pos, int flags, int add)
{
    char flagstr[512];
    char addstr[2] = { 0 };
    int trys = 0;

    RESTART_SPOT
    
    if (add)
	addstr[0] = '+';
    else
	addstr[0] = '-';
	    

    flagstr[0] = '\0';
    if (flags & FLAG_DELETE) {
      if ( flagstr[0] )
        strcat ( flagstr, " " );
	strcat(flagstr,"\\Deleted");
         /* JCS - Set this flag so that we can ask if the user
          * wants to expunge (delete from server) messages when
          * leaving this folder 
          */
	g->deletedMessages = TRUE;
    }
    if (flags & FLAG_SEEN) {
      if ( flagstr[0] )
        strcat ( flagstr, " " );
      strcat(flagstr,"\\Seen");
    }
    if (flags & FLAG_FLAGGED) {
      if ( flagstr[0] )
        strcat ( flagstr, " " );
      strcat(flagstr,"\\Flagged");
    }
    if (flags & FLAG_ANSWERED) {
      if ( flagstr[0] )
        strcat ( flagstr, " " );
      strcat(flagstr,"\\Answered");
    }

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "STORE %d %sFLAGS.SILENT (%a)",pos,addstr,flagstr);
    
    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) return -1;

    return 0;
}

int do_setflag(globals_t *g, int pos, int flags)
{
    return do_flag(g, pos,flags,1);
}
int do_removeflag(globals_t *g, int pos, int flags)
{
    return do_flag(g, pos,flags,0);
}

int do_mark_all_seen(globals_t *g)
{
    int trys = 0;

    RESTART_SPOT

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "STORE 1:* +FLAGS.SILENT (\\Seen)");
    
    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) return -1;

    return 0;
}

int do_createbboard(globals_t *g, const char *mboxname)
{
    int trys = 0;

    RESTART_SPOT

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "CREATE %s",mboxname);
    
    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Create of bboard failed");
	return -1;
    }
    
    return 0;
}

int do_deletebboard(globals_t *g, const char *mboxname)
{
    int trys = 0;

    RESTART_SPOT
    
    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "DELETE %s",mboxname);

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Delete of bboard failed");
	return -1;
    }

    return 0;
}

int do_renamebboard(globals_t *g, const char *oldmboxname, const char *newmboxname)
{
    int trys = 0;

    RESTART_SPOT

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "RENAME %s %s",oldmboxname,newmboxname);
    
    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Rename of bboard failed");
	return -1;
    }

    return 0;
}

#if 0
static char *
make_good_newlines(TEXT_BUF *tbuf, int *len)
{
    char buf[2048] = { 0 };
    int bytes_read;
    char *ret = malloc(tbuf->fsize*2+1);
    int tot_read = 0;
    int seenr = 0;
    if (!ret) return NULL;

    *len = 0;

    do {

	bytes_read = TXT_GetData(tbuf,
				 buf, sizeof(buf)-1, 
				 tot_read);

	if (bytes_read > 0) {
	    int i;

	    for (i = 0; i < bytes_read; i++) {
		switch(buf[i]) {
		case '\r':
		    seenr = 1;
		    /*
		     * Require \r\n. Covert bare \r to space
		     */
		    if (buf[i+1] != '\n') {
			buf[i] = ' ';
		    }
		    break;
		case '\n':
		    if (!seenr) {
			ret[(*len)++] = '\r';
		    }
		    seenr = 0;
		    break;
		default:
		    seenr = 0;
		    break;
		}
		ret[(*len)++] = buf[i];
	    }

	    tot_read+=bytes_read;

	} else {
	    free(ret);
	    return NULL;
	}

    } while (tot_read < tbuf->fsize);

    ret[(*len)] = '\0';

    if (*len != strlen(ret)) {
	printf("len doesn't equal string length\n");
    }

    return ret;
}
#endif
static char *
make_good_newlines ( TEXT_BUF * tbuf,
		     int      * len )
{
  char buf[LINE_SIZE + 1] = { 0 };
  int  bytes_read = 0;
  char * ret = xmalloc ( tbuf->fsize * 2 + 1 );
  int tot_read = 0;
  char * lastSpace;
  char * nextNewline;
  

  if ( !ret ) 
    return NULL;

  *len = 0;
  
  do { 

    bytes_read = TXT_GetData ( tbuf,
			       buf,
			       LINE_SIZE,
			       tot_read );
    buf[bytes_read] = '\0';

    if ( bytes_read > 0 ) { 
      
      nextNewline = strchr ( buf, '\n' );
      if ( nextNewline != NULL ) { 
	/* Partial line here */
	memcpy ( ret + (*len),
		 buf,
		 nextNewline + 1 - buf );
	tot_read += nextNewline + 1 - buf;
	*len += nextNewline + 1 - buf;

	if ( ret[(*len)-1] != '\r' ) {
	  ret[(*len)-1] = '\r';
	  ret[(*len)]  = '\n';
	  (*len) = (*len) + 1;
	}

      } else { 
	
	lastSpace = strrchr ( buf, ' ' );
	if ( lastSpace != NULL ) { 
	  memcpy ( ret + (*len),
		   buf, 
		   lastSpace + 1 - buf );
	  (*len) += lastSpace - buf;
	  tot_read += lastSpace - buf + 1;
	  ret[*(len)] = '\r';
	  ret[*(len)+1] = '\n';
	  (*len) = (*len) + 1;
	} else { 
	  
	  /* It's a whole line of characters! */
	  memcpy ( ret + (*len),
		   buf, 
		   LINE_SIZE - 1 );
	  tot_read += LINE_SIZE - 1;      
	  (*len) += LINE_SIZE - 1;
	  ret[(*len)] = '\r';
	  ret[(*len)+1] = '\n';
	  (*len) = (*len) + 2;
	}
      }
    } else { 
      free ( ret );
      return NULL;
    }


  } while ( tot_read  < tbuf->fsize );
  
  ret[(*len)] = '\0';
  return ret;
}

int do_append(globals_t *g, const char *mboxname, TEXT_BUF *tbuf)
{
    int len;
    char *buf;
    int trys = 0;

    if (!g) return -1;

    RESTART_SPOT

    buf = make_good_newlines(tbuf, &len);

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "APPEND %s (\\Seen) %s",mboxname, buf);

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }
	    	    
    free(buf);

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) return -1;

    return 0;
}

int do_draft(globals_t *g, const char *mboxname, TEXT_BUF *tbuf)
{
    int len, qpBufLen;
    char *buf, *qpBuf, *bufStart, *qpBufStart;
    int trys = 0;

    if (!g) return -1;

    RESTART_SPOT

    buf = make_good_newlines(tbuf, &len);
    qpBuf = xmalloc ( len * 2 + 1 );
    if ( qpBuf == NULL ) { 
      DLG_Alert ( ALERT_warning,
		  "Error: not enough memory to perform quoted-printable conversion" );
      free ( buf );
      return -1;      
    }
    
    /* We cannot encode the header as quoted-printable, so go 
     * past that.
     */
    bufStart = strstr ( buf, "\r\n\r\n" );
    if ( bufStart == NULL ) { 
      free ( buf );
      free ( qpBuf );
      CON_WriteLine ( "GAK! I could not find a newline ");
      return -1;
    }
    bufStart += 2;
    
    strncpy ( qpBuf, buf, bufStart - buf );

    qpBufStart = rfc2045_mime_headers ( MIME_TYPE_TEXT,
					"plain",
					NULL,
					"quoted-printable",
                                        NULL,
                                        RFC2183_UNKNOWN,
                                        NULL,
                                        NULL,
					qpBuf + ( bufStart - buf ),
					len * 2 + 1 - ( bufStart - buf ) );

    qpBufLen = rfc2045_to_mime_quoted_printable ( bufStart,
						  qpBufStart,
						  len * 2 + 1 - ( qpBufStart - qpBuf ) );
    if ( qpBufLen <= 0 ) { 
      DLG_Alert ( ALERT_warning,
		  "Error: could not encode in quoted-printable format" );
      free ( buf );
      free ( qpBuf );
      return -1;
    }

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "APPEND %s (\\Draft) %s", mboxname, qpBuf);

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }
	    	    
    free(buf);
    free(qpBuf);

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) return -1;

    return 0;
}

int do_expunge(globals_t *g)
{
    int trys = 0;

    /* xxx make sure a mailbox is open */
    
    RESTART_SPOT

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "EXPUNGE");

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Expunge failed");
	return -1;
    }

    return 0;
}

int do_fileinto(globals_t *g, int num, char *mbox)
{
    int trys = 0;
    /* xxx make sure a mailbox is open */

    RESTART_SPOT
    
    num++;

    g->cmd_done = NOTFINISHED;

    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "COPY %d %s",num,mbox);

    while (g->cmd_done == NOTFINISHED) {
        imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done == IMAP_EOF) goto restart;

    if (g->cmd_done!=IMAP_OK) {
	CON_WriteLine("Fileinto failed");
	return -1;
    }

    return 0;
}

static void do_checkfornewmail(globals_t *g)
{
    int unseen = 0;

    CON_SetBusy ( TRUE );

    imclient_addcallback(g->imclient_conn,
			 "STATUS", 0, callback_status_unseen,
			 (void *)&unseen, (char *)0);
    
    g->cmds_finished = 0;
    g->cmd_done = NOTFINISHED;
    imclient_send(g->imclient_conn, callback_finish, (void *)g,
		  "STATUS INBOX (UNSEEN)");
    
    while (g->cmd_done == NOTFINISHED) {
	imclient_processoneevent(g->imclient_conn);
    }

    if (g->cmd_done!=IMAP_OK) CON_WriteLine("Error checking for new mail");

    g->inboxunseen = unseen;

    CON_SetBusy ( FALSE );
}

static char rot13_char(char ch)
{
    if ((ch >= 'a') && (ch<='z')) {
	return ((ch - 'a' + 13) % 26) + 'a';
    }

    if ((ch >= 'A') && (ch<='Z')) {
	return ((ch - 'A' + 13) % 26) + 'A';
    }

    return ch;
}

static void rot13_line(char *line)
{
    int len;
    int j;

    len = strlen(line);
    for (j=0;j<len;j++) {
	line[j] = rot13_char(line[j]);
    }
}

int do_switch_rot13(void)
{
    int i;

    for (i=0;i<cur_bodyline;i++) {
	char line[100];

	RL_read(BODY_RL,i,line);

	rot13_line(line);

	RL_set(BODY_RL,i, line,0,NULL);
	LST_InvalidateItem(BODY_RL, i);
    }

    g_rot13_mode = (g_rot13_mode+1)%2;

    return 0;
}

int do_switch_fullheaders(int spot)
{
    g_fullheaders_mode = (g_fullheaders_mode+1)%2;

    return do_getmsg(do_getcurrentg(), spot);

    return 0;
}

#define CHECKMAILTIME 60*1
int g_idletime;

void do_events(select_info_pt si)
{
  extern int g_MayWaitRefresh;

  g_MayWaitRefresh = TRUE;

  /* could get incoming imap/smtp */
  

  g_MayWaitRefresh = FALSE;
  SCR_UpdateScreen();

  errno=0;

  si->nfound=select(99,
		   &si->read_fd,&si->write_fd,&si->except_fd,&si->timeout);

  if (si->nfound<0) {
    if (errno != EINTR) rl_error("select failed");
  } else if (si->nfound == 0) {
      int i;
      
      /* check for new mail */
      if ( si->timeout.tv_sec == 0 &&
	   si->timeout.tv_usec == 0 )
	si->timeout.tv_sec = 1;

      g_idletime += si->timeout.tv_sec;
      if (g_idletime >= CHECKMAILTIME) {
     
	  for (i = 0; i < num_globs; i++) {
	      char msg[100];
	      int oldinboxunseen;

	      globals_t *g = globs[i];
	      oldinboxunseen = g->inboxunseen;
	      
	      do_checkfornewmail(g);
	      
	      if ((g->inboxunseen > -1) && (oldinboxunseen > -1) && (g->inboxunseen > oldinboxunseen)) {
		  snprintf(msg,sizeof(msg)-1,"You have %d new message in your inbox (on server %s)",g->inboxunseen - oldinboxunseen,g->server_hostname);
		  DLG_NewMail(msg);
	      }
	  }

	  g_idletime = 0;
      }
  } else { 
    g_idletime = 0;
  }

}
