/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * common definitions to all mail client modules
 */

#ifndef COMMON_H
#define COMMON_H

#include <assert.h>
#ifdef IBMPC
#define RL_WINDOW_SIZE 50
#else
#define RL_WINDOW_SIZE 150
#endif

#define LINE_LENGTH 80

#ifdef AFS_ENV
extern int errno;
#endif

#include "msgs.h"

#ifndef MALLOC_DEFINED
extern char *malloc(), *realloc();
#endif

extern char  *index();

#if defined(mips) && defined(pmax_3)
extern char *memset();
#endif

/* define this, so only the front window has the highlight */

#ifndef DO_HIGHLIGHT_SELECTION
#define DO_HIGHLIGHT_SELECTION
#endif

typedef unsigned char Boolean;

#ifndef FALSE
#define FALSE	0
#endif

#ifndef TRUE
#define TRUE	1
#endif

#ifndef	NULL
#define NULL	0
#endif

#ifndef NIL
#define NIL	(char *)NULL
#endif

#define WRAP_INDEX(idx, max) (idx+1 >= max ? 0 : idx+1)

#define BWRAP_INDEX(idx, max) (idx-1 < 0 ? max-1 : idx-1)

#ifndef imin
#define imin(xx1,xx2) (((xx1)<(xx2))?(xx1):(xx2))
#define imax(xx1,xx2) (((xx1)>(xx2))?(xx1):(xx2))
#define iabs(xx1)     (((xx1)<0)?(-(xx1)):(xx1))
#endif

#define LINE_WIDTH  80

#define NUM_LOGIN_FIELDS 3
#define LOGIN_USER 0
#define LOGIN_PWD  1
#define LOGIN_IMAP 1
#define LOGIN_SMTP 2

#define GUARDIAN_PORT 900

extern char userid[], imap_host[], smtp_host[], password[];

#define STR255LEN 255

#define MAX_BAN STR255LEN


typedef struct remote_prefs_struct {
  Boolean valid;
  Boolean bools[8];
  char str_outgoing[STR255LEN];
  char str_remote_printer[STR255LEN];
  char str_unused1[STR255LEN];
  char str_unused2[STR255LEN];
} REMOTE_PREFS;

#define BOOL_PREFS_BSIGN 0
#define BOOL_PREFS_BPURG 1
#define BOOL_PREFS_BCOPY 2

#define REM_PREFS_BSIGN 1
#define REM_PREFS_BCOPY 2
#define REM_PREFS_SCOPY 3
#define REM_PREFS_SPRNT 4
#define REM_PREFS_BPURG 5
#define REM_PREFS_HEADS 6
#define REM_PREFS_FAXIT 7
#define REM_PREFS_PUNTREAD 8


#define NULL_INT_FUNC_pt  (int  (*)())NULL
#define NULL_VOID_FUNC_pt (void (*)())NULL
#define NULL_LONG_FUNC_pt (long (*)())NULL
#define NULL_CHAR_FUNC_pt (char (*)())NULL

extern long FIO_GetFileSize();

#define OPEN_R  (0x01) /* open in binary mode */
#define OPEN_W  (0x02) /* open in binary mode */
#define OPEN_A  (0x04) /* open in binary mode */

#ifdef IBMPC
#define OPEN_RT (0x08) /* open in text mode, cr/lf -> lf conversion */
#define OPEN_WT (0x10) /* open in text mode, cr/lf -> lf conversion */
#define OPEN_AT (0x20) /* open in text mode, cr/lf -> lf conversion */
#else  /* IBMPC */
#define OPEN_RT OPEN_R
#define OPEN_WT OPEN_W
#define OPEN_AT OPEN_A
#endif /* IBMPC */

#define OPEN_RW (OPEN_R | OPEN_W)

/* Mail forwarding subcommands */

#define MFD_SET   -1
#define MFD_QUERY -2
#define MFD_CLEAR -3

#define FAX_INFO_NAME    0
#define FAX_INFO_ADDR1   1
#define FAX_INFO_ADDR2   2
#define FAX_INFO_ADDR3   3
#define FAX_INFO_PHONE   4
#define FAX_INFO_VOICE   5
#define FAX_INFO_CHARGE  6
#define FAX_INFO_TIME    7
#define FAX_INFO_SUBJECT 8

#define FAX_OPT_COVER    0
#define FAX_OPT_BODY     1
#define FAX_OPT_RETRY    2

#define FAX_OPT_ACK      0
#define FAX_OPT_RES      1
#define FAX_OPT_CONF     2
#define FAX_OPT_TRAIL    3

typedef struct fax_headers_R {
  char fax_destinfo[FAX_INFO_SUBJECT + 1][STR255LEN];
  char fax_options[FAX_OPT_RETRY + 1][STR255LEN];
  Boolean fax_bools[FAX_OPT_TRAIL+1];
} FAX_HEADERS, *FAX_HEADERS_pt;

extern FAX_HEADERS_pt g_FaxHeaders;

#define TF_CLEAR(xx, yy)     ((xx) &= (~(yy)))
#define TF_SET(xx, yy)       ((xx) |= (yy))
#define TF_ISSET(xx, yy)     ((xx) & (yy))
#define TF_ALL_ISSET(xx, yy) (((xx) & (yy)) == (yy))

extern Boolean gWeAreBusy;
extern Boolean gShowFullHeaders;
extern Boolean gFilerLongList;
extern Boolean gCanSendFax;
extern Boolean gSignMail;
extern Boolean gBccMe;
extern Boolean gNovice;
extern Boolean gPuntRead;
extern Boolean gCapturedShell;

extern int     g_MayWaitRefresh;

extern char *KeyMapFilename;
extern char *KeyMapDirectory;

typedef struct point_struct {
  unsigned int x, y;
} POINT;

#define SCREEN_BACKGROUND  0

#ifndef IBMPC
extern int errno;
#endif /* IBMPC */

int want_busy_command(void);

void HintUser(char *msg,int keytab[], int keytabsize);

void clear_checks(void);

void set_net_is_connecting(int aval);

int DoMenuCommand(int menuid);

int get_local_string(int id,char *buf);

void ack_bytes(long apos);

void get_boolean_from_user(char *prompt,int default_item, int lii);

void get_string_from_user(char *prompt,int lii);

void choose_from_list(char *prompt,int default_item, int lii);

void DisplayServerHint(char *hint_msg);

void close_cleanup(void);

#endif /* COMMON_H */
