/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * rlci functions in login segment
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <strings.h>

#include <rlci.h>

long bytes_transmitted;

/*
 * initilize client remote lists
 */

void rlci_initialize()
{
  static int RLCI_has_initialized = FALSE;
  char *saved_buf;
  int i;

  the_end_is_comming=0;

  for (i=0; i<RL_NUM_LISTS; i++) {
    saved_buf = RL_all[i].data;
    bzero(&RL_all[i], sizeof(RL_header));
    RL_all[i].lii = i;
#ifndef ALLOCATE_CONSOLE_BUFFER
    if (i == CONSOLE_RL) continue;
#endif /* ALLOCATE_CONSOLE_BUFFER */
    if (RLCI_has_initialized)
      RL_all[i].data = saved_buf;
    else 
      RL_all[i].data = c_get_mem(rl_chunk_size * RL_LENGTH);
  }
  RLCI_has_initialized = TRUE;
}

void guardian_startup(user,pwd,service,vers)
char *user;
char *pwd;
char *service;
char *vers;
{
    int aloc_siz;
    int servlen;
    char servbuf[40];
    int plen=strlen(pwd);
    int ulen=strlen(user);
    int verslen=strlen(vers);

    if(strlen(service)>(sizeof(servbuf)-10))
	rl_error("service name too long");

    sprintf(servbuf,"%s.%d.%d",service,RLI_MAJOR_VERSION,RLI_MINOR_VERSION);
    servlen=strlen(servbuf);
    aloc_siz=30+ulen+plen+servlen+verslen;
  
    /*    PFpack(
	&ppfm,
         LS_aloc(aloc_siz),
         LS_msend_num(LS_int16, 8),
         LS_msend_num(LS_int16, 9),
	 LS_msend_str(ulen,user),
	 LS_msend_str(plen,pwd),
	 LS_msend_str(servlen,servbuf),
	 LS_msend_str(verslen,vers),
         LS_dropoff); */
}

extern char *getvstr();
extern char *getvstr_cs();

void data_startup(cons_port)
int cons_port;
{
#ifdef IBMPC
    char *vstr=getvstr();
#else
    char *vstr=getvstr_cs();
#endif
    int verslen;
    verslen=strlen(vstr);

    bytes_transmitted = 0L;

    /*    PFpack(
	&ppfm,
         LS_aloc(40),
         LS_seqophead(CONSOLE_RL, 0, SYSM_MASK+SYSM_hello),
         LS_msend_num(LS_int16, 1),	
         LS_msend_num(LS_int16, RLI_MAJOR_VERSION),
         LS_msend_num(LS_int16, RLI_MINOR_VERSION),
         LS_msend_num(LS_int16, cons_port),
	 LS_msend_str(verslen,vstr),
         LS_msend_num(LS_int32,rl_chunk_size),
         LS_dropoff); */



}

void set_rl_window_size(int requested_rl_size, int screen_size)
{
  rl_chunk_size = imax(requested_rl_size, (2 * screen_size));
  half_buf = rl_chunk_size / 2;
}
