/* 
 * Copyright (c) 2000 Carnegie Mellon University.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The name "Carnegie Mellon University" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For permission or any other legal
 *    details, please contact  
 *      Office of Technology Transfer
 *      Carnegie Mellon University
 *      5000 Forbes Avenue
 *      Pittsburgh, PA  15213-3890
 *      (412) 268-4387, fax: (412) 268-7395
 *      tech-transfer@andrew.cmu.edu
 *
 * 4. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by Computing Services
 *     at Carnegie Mellon University (http://www.cmu.edu/computing/)."
 *
 * CARNEGIE MELLON UNIVERSITY DISCLAIMS ALL WARRANTIES WITH REGARD TO
 * THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE
 * FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SCREEN_H
#define SCREEN_H

/*
 * screen handling attributes
 */

#define NORMAL    0                 /* Normal on all displays           */
#define HILITE    1                 /* Hilite on all displays           */
#define RVIDEO    2                 /* Reverse Video                    */
#define BLINKH    3                 /* Blink + HILITE                   */
#define REGDIF    4                 /* RVIDEO unless B/W adapter        */
#define NORMCE    5                 /* Same as NORMAL, unless color     */
#define NORMCI    6                 /* Same as NORMAL, unless color     */
#define HILITC    7                 /* Same as HILITE, unless color     */
#define RVIDHI    8                 /* Reverse Video + HILITE           */
#define WHTBLU    9                 /* White on blue, or RVIDEO         */
#define INVISI    10                /* Invisible                        */
#define RINVIS    11                /* Reverse Video & invisible data   */
#define UNDERN    12                /* Normal underline                 */
#define UNDERH    13                /* Hilited underline                */
#define GRAFIC    14                /* Use graphic characters           */
#define WHTGRY    15                /* White on Grey                    */
#define BLKGRY    16                /* Black on Grey                    */
#define NORMDM    17                /* DIM if color, else normal        */
#define RVDGRY    18                /* Reverse video on grey            */
#define BWHITE    19                /* bright white (color),else hilite */
#define DWHITE    20                /* dim white (color), else normal   */

#define BAN_H_ON  RVIDEO
#define BAN_H_OFF HILITE

#define DOUBLE_LINE_BOX 0
#define SINGLE_LINE_BOX 1

/*
 * PC color definitions CGA/EGA/VGA
 *
 *  7 6 5 4 3 2 1 0
 *  1 . . . . . . .   Blinking / intensity of background
 *  . 1 . . . . . .   Red background
 *  . . 1 . . . . .   Green background
 *  . . . 1 . . . .   Blue background
 *  . . . . 1 . . .   Foreground intensity
 *  . . . . . 1 . .   Red foreground
 *  . . . . . . 1 .   Green foreground
 *  . . . . . . . 1   Blue foreground
 *
 */

#define BLACK     0x00
#define BLUE      0x01
#define GREEN     0x02
#define CYAN      0x03
#define RED       0x04
#define MAGENTA   0x05
#define BROWN     0x06
#define GRAY      0x07
#define DARKGRAY  0x08
#define HBLUE     0x09
#define HGREEN    0x0A
#define HCYAN     0x0B
#define HRED      0x0C
#define HMAGENTA  0x0D
#define HYELLOW   0x0E
#define HWHITE    0x0F


typedef struct videoparms {
    unsigned char palette;
    unsigned char page;
    unsigned char attr;
    unsigned char mode;
    unsigned char columnss;
} VIDEOPARMS;

#define MONO_ATTR_TABLE \
    0x07, 0x0F, 0x70, 0x8F, 0x70, 0x07, 0x07, \
    0x0F, 0x78, 0x70, 0x00, 0x70, 0x01, 0x09, \
    0x07, 0x0F, 0x70, 0x07, 0x70, 0x0F, 0x07

#define BW_ATTR_TABLE \
    0x07, 0x0F, 0x70, 0x8F, 0x0F, 0x07, 0x07, \
    0x0F, 0x78, 0x70, 0x00, 0x70, 0x07, 0x0F, \
    0x07, 0x0F, 0x70, 0x07, 0x70, 0x0F, 0x07

#define COLOR_ATTR_TABLE \
    0x0B, 0x0E, 0x4F, 0x84, 0x4F, 0x0C, 0x0A, \
    0x0D, 0x4E, 0x17, 0x00, 0x4F, 0x0B, 0x0E, \
    0x0B, 0x0F, 0x3B, 0x03, 0x17, 0x0F, 0x07
/*  0x0B, 0x3F, 0x3B, 0x03, 0x17, 0x0F, 0x07 */

#define EGACO_ATTR_TABLE \
    0x0B, 0x0E, 0x4F, 0x84, 0x4F, 0x0C, 0x0A, \
    0x0D, 0x4E, 0x17, 0x00, 0x4F, 0x01, 0x09, \
    0x0B, 0x3F, 0x3B, 0x03, 0x17, 0x0F, 0x07

#define TERMCAP_ATTR_TABLE \
    0, 0, 1, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0

extern unsigned char screen_rows, screen_cols;

typedef struct box_struct {
  POINT loc;
  POINT dim;
} BOX;

typedef struct win_struct {
  int id;
  POINT loc;
  POINT dim;
  POINT zoomloc;
  POINT zoomdim;
  int   data_height;
  int   data_width;
  Boolean zoom;
  Boolean shown;
  int   window_type;
  void (*hproc)();
  int  (*dokey)();
  void (*draw)();
  void (*close_proc)();
  char title[MAX_BAN];
} WINFO;

typedef struct win_list_struct {
  WINFO *awin;
  struct win_list_struct *prev;
  struct win_list_struct *next;
} WINLIST;

extern WINLIST *win_list;

extern WINLIST *WIN_Create();

extern char *SCR_GetKeyText();

extern Boolean g_WinRegionRefresh;

#define GET_FrontWindow win_list

#define WindowZoomed(xx_win) ((xx_win) && (xx_win)->awin->zoom)

#define WTYPE_LIST 0x01
#define WTYPE_TEXT 0x02
#define WTYPE_DLOG 0x04

#define WIN_FrontWindowType ((GET_FrontWindow)->awin->window_type)

#define WIN_FrontWindowTitle ((GET_FrontWindow)->awin->title)

#define WIN_Show(x)  (x)->awin->shown = TRUE
#define WIN_Hide(x)  (x)->awin->shown = FALSE
#define WIN_Focus(x) ((x)->awin->focus)
#define WIN_isShown(x) ((x)->awin && (x)->awin->shown != 0)

#define BAN_ROW 1
#define BAN_COL 1
#define DATA_ROW 3
#define DATA_COL 1
#define HEIGHT_OFFSET 4
#define WIDTH_OFFSET 2

extern int w_ban_row, w_ban_col, w_data_row, w_data_col;
extern int w_height_offset, w_width_offset, w_mode;

#define W_NORMAL_MODE  0x01
#define W_ZOOM_ALL     0x02
#define W_NOBAN_MODE   0x04
#define W_NOBORD_MODE  0x08

/* graphical characters for drawing lines/boxes */

#define NO_C_SET 0
#define VT_C_SET 1
#define ZT_C_SET 2

#define VT_HL 'q'
#define VT_VL 'x'
#define VT_LR 'j'
#define VT_UR 'k'
#define VT_UL 'l'
#define VT_LL 'm'
#define VT_PL 'u'
#define VT_PR 't'
#define VT_ST 'o'
#define VT_SB 's'

#define H19_HL 'a'
#define H19_VL '\140'
#define H19_LR 'd'
#define H19_UR 'c'
#define H19_UL 'f'
#define H19_LL 'e'
#define H19_PL 't'
#define H19_PR 'v'
#define H19_ST 'z'
#define H19_SB '{'

#define MIN_WIDTH  10

#define Background(xx_color) ((xx_color << 4) & 0x7F)

#endif /* SCREEN_H */
