/***********************************************************
        Copyright 1998 by Carnegie Mellon University

                      All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of CMU not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.

CMU DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
CMU BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.
******************************************************************/

#ifndef SKIP_LIST_H
#define SKIP_LIST_H

/* this is a simple implementation of skiplists.  it takes void *
   pointers, and a int comp(void *, void *) comparison function. */

/* JCS - increased this from 30, because it was wayyyyyyy too small */
#define SKIP_ABSMAXLVL 4096

typedef struct Skipnode {
    void *data;
    struct Skipnode *forward[1];
} skipnode;

typedef struct Skiplist {
    int maxlevel;
    float prob;
    int curlevel;
    int items;
    int (*comp)(const void *, const void *);
    skipnode *header; /* phantom first element of skiplist */
    skipnode *finger[SKIP_ABSMAXLVL]; /* finger for localization */
} skiplist;

/* ml = maximum level of each node
   p = probability of next higher level
   cf = comparator function for elements */
skiplist *newlist(int ml, float p,  int (*cf)(const void *, const void *));
void *ssearch(skiplist *S, void *k);
void sinsert(skiplist *S, void *k);
void sdelete(skiplist *S, void *k);
void destroylist(skiplist *S);
int sempty(skiplist *S);
void sforeach(skiplist *S, void (*f)(const void *));

void *sfirst(skiplist *S, skipnode **ptr);
void *snext(skipnode **ptr);

#endif
